from enum import Enum


class BiomeEnum(Enum):
    dwelling = "cave"
    jungle = "jungle"
    volcana = "volcano"
    tide_pool = "tidepool"
    temple_of_anubis = "temple"
    ice_caves = "ice"
    neo_babylon = "babylon"
    sunken_city = "sunken"
    eggplant_world = "eggplant"
    surface = "surface"


class BossBiomeEnum(Enum):
    olmecs_lair = "????"  # maybe it doesn't count as a biome?
    abzu = "abzu"
    duat = "duat"
    city_of_gold = "?????"
    tiamats_throne = "?????"
    hunduns_hideaway = "????"
    cosmic_ocean = "cosmic"



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional
from typing import TextIO

from .level_chances import LevelChances, LevelChance
from .level_settings import LevelSettings, LevelSetting
from .level_templates import LevelTemplates, LevelTemplate
from .monster_chances import MonsterChances, MonsterChance
from .tile_codes import TileCodes, TileCode
from .utils import DirectivePrefixes, Peekable

SECTION_COMMENT = "// ------------------------------"


def parse_section_comment(line, file_handle):
    output = f"{line}\n"

    next_line = file_handle.peek()
    while next_line and next_line.startswith("//"):
        output += next_line

        # Advance the cursor and peek the next line.
        file_handle.advance()
        next_line = file_handle.peek()

    return output


@dataclass
class LevelFile:
    comment: Optional[str]
    level_settings: LevelSettings
    tile_codes: TileCodes
    level_chances: LevelChances
    monster_chances: MonsterChances
    level_templates: LevelTemplates

    @classmethod
    def from_path(cls, level_path: Path) -> "LevelFile":
        level_file = LevelFile(
            comment=None,
            level_settings=LevelSettings(),
            tile_codes=TileCodes(),
            level_chances=LevelChances(),
            monster_chances=MonsterChances(),
            level_templates=LevelTemplates(),
        )

        last_section_comment = None
        last_seen_directive = None

        with level_path.open("r") as level_fh:
            level_fh = Peekable(level_fh)
            for line in level_fh:

                line = line.strip()
                if not line:
                    continue

                if line.startswith(DirectivePrefixes.LEVEL_SETTING.value):
                    level_setting = LevelSetting.parse(line)
                    level_file.level_settings.set_obj(level_setting)
                    if last_seen_directive != DirectivePrefixes.LEVEL_SETTING:
                        if last_section_comment:
                            level_file.level_settings.comment = last_section_comment
                            last_section_comment = None
                        last_seen_directive = DirectivePrefixes.LEVEL_SETTING

                elif line.startswith(DirectivePrefixes.TILE_CODE.value):
                    tile_code = TileCode.parse(line)
                    level_file.tile_codes.set_obj(tile_code)
                    if last_seen_directive != DirectivePrefixes.TILE_CODE:
                        if last_section_comment:
                            level_file.tile_codes.comment = last_section_comment
                            last_section_comment = None
                        last_seen_directive = DirectivePrefixes.TILE_CODE

                elif line.startswith(DirectivePrefixes.LEVEL_CHANCE.value):
                    level_chance = LevelChance.parse(line)
                    level_file.level_chances.set_obj(level_chance)
                    if last_seen_directive != DirectivePrefixes.LEVEL_CHANCE:
                        if last_section_comment:
                            level_file.level_chances.comment = last_section_comment
                            last_section_comment = None
                        last_seen_directive = DirectivePrefixes.LEVEL_CHANCE

                elif line.startswith(DirectivePrefixes.MONSTER_CHANCE.value):
                    monster_chance = MonsterChance.parse(line)
                    level_file.monster_chances.set_obj(monster_chance)
                    if last_seen_directive != DirectivePrefixes.MONSTER_CHANCE:
                        if last_section_comment:
                            level_file.monster_chances.comment = last_section_comment
                            last_section_comment = None
                        last_seen_directive = DirectivePrefixes.MONSTER_CHANCE

                elif line.startswith(DirectivePrefixes.TEMPLATE.value):
                    template = LevelTemplate.parse(line, level_fh)
                    level_file.level_templates.set_obj(template)
                    if last_seen_directive != DirectivePrefixes.TEMPLATE:
                        if last_section_comment:
                            level_file.level_templates.comment = last_section_comment
                            last_section_comment = None
                        last_seen_directive = DirectivePrefixes.TEMPLATE

                elif line == SECTION_COMMENT:
                    last_section_comment = parse_section_comment(line, level_fh)
                    if not level_file.comment and not last_seen_directive:
                        level_file.comment = last_section_comment
                        last_section_comment = None

        return level_file

    def write(self, handle: TextIO):
        handle.write(f"{self.comment}\n")
        self.level_settings.write(handle)
        self.tile_codes.write(handle)
        self.level_chances.write(handle)
        self.monster_chances.write(handle)
        self.level_templates.write(handle)

    def write_path(self, level_path: Path):
        with level_path.open("w") as level_fh:
            self.write(level_fh)

    def print(self):
        self.write(sys.stdout)



import argparse
import logging
from pathlib import Path

from .assets import AssetStore
from .constants import (
    DEFAULT_COMPRESSION_LEVEL,
    EXTRACTED_DIR,
    FILEPATH_DIRS,
    OVERRIDES_DIR,
    PACKS_DIR,
)

DEFAULT_MODS_DIR = "Mods"

TOP_LEVEL_DIRS = [EXTRACTED_DIR, PACKS_DIR, OVERRIDES_DIR]


def main():
    parser = argparse.ArgumentParser(description="Extract Spelunky 2 Assets.")
    parser.add_argument("exe", type=argparse.FileType("rb"), help="Path to Spel2.exe")
    parser.add_argument(
        "--mods-dir",
        type=str,
        default=DEFAULT_MODS_DIR,
        help="Path to directory containing mods. Default: %(default)s",
    )
    parser.add_argument(
        "--extracted-name",
        type=str,
        default=EXTRACTED_DIR,
        help="Name of the directory to extract files into. Default: %(default)s",
    )
    parser.add_argument(
        "--compression-level",
        type=int,
        default=DEFAULT_COMPRESSION_LEVEL,
        help=(
            " Value between 1 and 22 (higher = smaller data size)"
            " - if modified assets are too large, increase compression"
        ),
    )
    parser.add_argument(
        "--no-recompress",
        dest="recompress",
        default=True,
        action="store_false",
        help=(
            "Don't recompress extracted assets. Increases extract time "
            "but packing will be slower."
        ),
    )
    parser.add_argument(
        "--create-entity-sheets",
        dest="create_entity_sheets",
        default=False,
        action="store_true",
        help=(
            "Create extended entity assets merged from multiple sheets."
        ),
    )
    parser.add_argument(
        "--no-mkdirs",
        dest="mkdirs",
        default=True,
        action="store_false",
        help="Don't make directories for Mods.",
    )

    args = parser.parse_args()

    mods_dir = Path(args.mods_dir)
    extracted_dir = Path(args.extracted_name)

    logging.basicConfig(format="%(levelname)s - %(message)s", level=logging.INFO)

    if args.mkdirs:
        # Make all directories for extraction and overrides
        for dir_ in TOP_LEVEL_DIRS:
            (mods_dir / dir_).mkdir(parents=True, exist_ok=True)

        for dir_ in FILEPATH_DIRS:
            (mods_dir / extracted_dir / dir_).mkdir(parents=True, exist_ok=True)
            (mods_dir / ".compressed" / extracted_dir / dir_).mkdir(
                parents=True, exist_ok=True
            )

    asset_store = AssetStore.load_from_file(args.exe)
    unextracted = asset_store.extract(
        mods_dir / extracted_dir,
        mods_dir / ".compressed" / extracted_dir,
        args.compression_level,
        recompress=args.recompress,
        create_entity_sheets=args.create_entity_sheets,
    )

    for asset in unextracted:
        logging.warning("Un-extracted Asset %s", asset)


if __name__ == "__main__":
    main()



from pathlib import Path

from .base_classes import BaseSpriteLoader


class JournalTrapSheet(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/journal_entry_traps.png")
    _chunk_size = 160
    _chunk_map = {
        "journal_arrow_trap": (0, 0, 1, 1),
        "journal_spear_trap": (1, 0, 2, 1),
        "journal_thorny_vine": (2, 0, 3, 1),
        "journal_powder_box": (3, 0, 4, 1),
        "journal_falling_platform": (4, 0, 5, 1),
        "journal_crush_trap": (6, 0, 7, 1),
        "journal_laser_trap": (7, 0, 8, 1),
        "journal_spring_trap": (8, 0, 9, 1),
        "journal_landmine": (9, 0, 10, 1),
        "journal_spikes": (0, 1, 1, 2),
        "journal_spark_trap": (1, 1, 2, 2),
        "journal_egg_sac": (2, 1, 3, 2),
        "journal_bear_trap": (3, 1, 4, 2),
        "journal_totem_trap": (0, 4, 1, 6),
        "journal_lion_trap": (1, 4, 2, 6),
        "journal_log_trap": (2, 4, 3, 6),
        "journal_sticky_trap": (3, 4, 4, 6),
        "journal_sliding_wall": (4, 4, 5, 6),
        "journal_giant_crush_trap": (0, 8, 2, 10),
        "journal_boulder": (2, 8, 4, 10),
        "journal_giant_clam": (4, 8, 6, 10),
        "journal_bone_drop": (8, 8, 10, 10),
        "journal_frog_trap": (6, 9, 8, 10),
    }



import sys
from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent
APP_DIR = BASE_DIR
ROOT_DIR = BASE_DIR.parent.parent
IS_EXE = False

# Setup static files to work with onefile exe
if hasattr(sys, "_MEIPASS"):
    BASE_DIR = BASE_DIR / getattr(sys, "_MEIPASS")
    APP_DIR = Path(sys.executable).resolve().parent
    ROOT_DIR = BASE_DIR
    IS_EXE = True


from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from PIL import Image

from .base_classes import AbstractBiome, BaseSpriteLoader, DEFAULT_BASE_PATH


class SpelukySpriteFetcher:
    def __init__(self, base_path: Path = DEFAULT_BASE_PATH):
        self.base_path = base_path
        # Setting up universal pieces
        self._non_biome_sheets, self._non_biome_map = self._make_non_biome_map()
        # Now biome specific pieces
        self._biome_dict = self._init_biomes()
        self._biome_map = {k: v.get for k, v in self._biome_dict.items()}

    def _init_biomes(self) -> Dict[str, AbstractBiome]:
        from . import biomes

        return {
            getattr(biomes, b).biome_name: getattr(biomes, b)(self.base_path)
            for b in biomes.__all__
        }

    def _make_non_biome_map(self) -> Tuple[List[BaseSpriteLoader], Dict[str, Callable]]:
        from . import monsters
        from .items import ItemSheet
        from .hud import HudSheet
        from .floormisc import FloorMiscSheet

        # Gather all of the sheets in a list, these are the classes, not instances yet
        sheets = [getattr(monsters, m) for m in monsters.__all__]
        sheets.extend([ItemSheet, HudSheet, FloorMiscSheet])
        # Now making them instances
        sheets = [s(self.base_path) for s in sheets]
        key_map = {}
        for sheet in sheets:
            for k in sheet.key_map():
                key_map[k] = sheet.get
        return sheets, key_map

    # noinspection PyNoneFunctionAssignment
    def get(self, name: str, biome: str = "cave") -> Optional[Image.Image]:
        # we are getting the right sprite sheet with the first `(name)` then getting the
        # image with the second `(name)` maybe using partials here would be more
        # pleasing?
        img = self._non_biome_map.get(name, lambda x: None)(name)
        if not img:
            img = self._biome_map.get(biome, lambda x: None)(name)
        if not img:
            for bname, b_class in self._biome_dict.items():
                if bname != biome:
                    img = b_class.get(name)
                if img:
                    break
        return img



import logging
import os
import os.path
import random
import tkinter as tk
import tkinter.font as font
import tkinter.messagebox as tkMessageBox
from tkinter import filedialog, ttk
import tempfile
from pathlib import Path

from PIL import Image, ImageDraw, ImageEnhance, ImageFont, ImageTk

from modlunky2.constants import BASE_DIR
from modlunky2.ui.widgets import LevelsTree, RulesTree, ScrollableFrame, Tab

logger = logging.getLogger("modlunky2")


class LevelsTab(Tab):
    def __init__(self, tab_control, install_dir, *args, **kwargs):
        super().__init__(tab_control, *args, **kwargs)
        self.last_selected_room = None
        # TODO: Get actual resolution
        self.screen_width = 1290
        self.screen_height = 720
        self.extracts_mode = True
        self.dual_mode = False
        self.tab_control = tab_control
        self.install_dir = install_dir
        self.textures_dir = install_dir / "Mods/Extracted/Data/Textures"

        self.lvl_editor_start_canvas = tk.Canvas(self)
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.lvl_editor_start_canvas.grid(row=0, column=0, columnspan=2, sticky="nswe")
        self.lvl_editor_start_canvas.columnconfigure(0, weight=1)
        self.lvl_editor_start_canvas.rowconfigure(0, weight=1)

        self.extracts_path = self.install_dir / "Mods" / "Extracted" / "Data" / "Levels"
        self.overrides_path = self.install_dir / "Mods" / "Overrides"

        self.welcome_label = tk.Label(
            self.lvl_editor_start_canvas,
            text=(
                "Welcome to the Spelunky 2 Level Editor! "
                "Created by JackHasWifi with lots of help from "
                "Garebear, Fingerspit, Wolfo, and the community\n\n "
                "NOTICE: Saving when viewing extracts will save the "
                "changes to a new file in overrides.\n"
                "When loading from extracts, if a file exists in overrides,\nit will be loaded from there instead.\n\n"
                "BIGGER NOTICE: Please make backups of your files. This is still in beta stages.."
            ),
            anchor="center",
            bg="black",
            fg="white",
        )
        self.welcome_label.grid(row=0, column=0, sticky="nswe", ipady=30, padx=(10, 10))

        def select_lvl_folder():
            dirname = filedialog.askdirectory(
                parent=self, initialdir="/", title="Please select a directory"
            )
            if not dirname:
                return
            else:
                self.extracts_mode = False
                self.lvls_path = dirname
                self.load_editor()

        def load_extracts_lvls():
            if os.path.isdir(self.extracts_path):
                self.extracts_mode = True
                self.lvls_path = self.extracts_path
                self.load_editor()

        self.btn_lvl_extracts = ttk.Button(
            self.lvl_editor_start_canvas,
            text="Load From Extracts",
            command=load_extracts_lvls,
        )
        self.btn_lvl_extracts.grid(
            row=1, column=0, sticky="nswe", ipady=30, padx=(20, 20), pady=(10, 1)
        )

        self.btn_lvl_folder = ttk.Button(
            self.lvl_editor_start_canvas,
            text="Load Levels Folder",
            command=select_lvl_folder,
        )
        self.btn_lvl_folder.grid(
            row=2, column=0, sticky="nswe", ipady=30, padx=(20, 20), pady=(10, 10)
        )

    def load_editor(self):
        self.save_needed = False
        self.last_selected_file = None
        self.tiles = None
        self.tiles_meta = None
        self.lvl_editor_start_canvas.grid_remove()
        self.columnconfigure(0, minsize=200)  # Column 0 = Level List
        self.columnconfigure(0, weight=0)
        self.columnconfigure(1, weight=1)  # Column 1 = Everything Else
        self.rowconfigure(0, weight=1)  # Row 0 = List box / Label

        # Loads lvl Files
        self.tree_files = ttk.Treeview(
            self, selectmode="browse"
        )  # This tree shows all the lvl files loaded from the chosen dir
        self.tree_files.place(x=30, y=95)
        self.vsb_tree_files = ttk.Scrollbar(
            self, orient="vertical", command=self.tree_files.yview
        )
        self.vsb_tree_files.place(x=30 + 200 + 2, y=95, height=200 + 20)
        self.tree_files.configure(yscrollcommand=self.vsb_tree_files.set)
        self.tree_files["columns"] = ("1",)
        self.tree_files["show"] = "headings"
        self.tree_files.column("1", width=100, anchor="w")
        self.tree_files.heading("1", text="Level Files")
        self.my_list = os.listdir(self.lvls_path)
        self.tree_files.grid(row=0, column=0, rowspan=1, sticky="nswe")
        self.vsb_tree_files.grid(row=0, column=0, sticky="nse")

        for (
            file
        ) in (
            self.my_list
        ):  # Loads list of all the lvl files in the left farthest treeview
            if str(file).endswith(".lvl"):
                self.tree_files.insert("", "end", values=str(file), text=str(file))

        # Seperates Level Rules and Level Editor into two tabs
        self.tab_control = ttk.Notebook(self)
        self.tab_control.grid(row=0, column=1, rowspan=3, sticky="nwse")

        self.tab1 = ttk.Frame(self.tab_control)  # Tab 1 shows a lvl files rules
        self.tab2 = ttk.Frame(self.tab_control)  # Tab 2 is the actual level editor

        self.button_back = tk.Button(
            self, text="Go Back", bg="black", fg="white", command=self.go_back
        )
        self.button_back.grid(row=1, column=0, sticky="nswe")

        self.button_save = tk.Button(
            self,
            text="Save",
            bg="purple",
            fg="white",
            command=self.save_changes,
        )
        self.button_save.grid(row=2, column=0, sticky="nswe")

        self.tab_control.add(  # Rules Tab ----------------------------------------------------
            self.tab1, text="Rules"
        )

        self.tab1.columnconfigure(0, weight=1)  # Column 1 = Everything Else
        self.tab1.rowconfigure(0, weight=1)  # Row 0 = List box / Label

        self.tree = RulesTree(
            self.tab1, selectmode="browse"
        )  # This tree shows rules parses from the lvl file
        self.tree.bind("<Double-1>", lambda e: self.on_double_click(self.tree))
        self.tree.place(x=30, y=95)
        # style = ttk.Style(self)
        self.vsb = ttk.Scrollbar(self.tab1, orient="vertical", command=self.tree.yview)
        self.vsb.place(x=30 + 200 + 2, y=95, height=200 + 20)
        self.tree.configure(yscrollcommand=self.vsb.set)
        self.tree["columns"] = ("1", "2", "3")
        self.tree["show"] = "headings"
        self.tree.column("1", width=100, anchor="w")
        self.tree.column("2", width=10, anchor="w")
        self.tree.column("3", width=100, anchor="w")
        self.tree.heading("1", text="Entry")
        self.tree.heading("2", text="Value")
        self.tree.heading("3", text="Notes")
        self.tree.grid(row=0, column=0, sticky="nwse")
        self.vsb.grid(row=0, column=1, sticky="nse")

        self.tab_control.add(  # Level Editor Tab -------------------------------------------------
            self.tab2, text="Level Editor"
        )
        self.tab2.columnconfigure(0, minsize=200)  # Column 0 = Level List
        self.tab2.columnconfigure(1, weight=1)  # Column 1 = Everything Else
        self.tab2.rowconfigure(2, weight=1)  # Row 0 = List box / Label

        self.tree_levels = ttk.Treeview(
            self.tab2, selectmode="browse"
        )  # This tree shows the rooms in the level editor
        self.tree_levels.place(x=30, y=95)
        self.vsb_tree_levels = ttk.Scrollbar(
            self.tab2, orient="vertical", command=self.tree_levels.yview
        )
        self.vsb_tree_levels.place(x=30 + 200 + 2, y=95, height=200 + 20)
        self.tree_levels.configure(yscrollcommand=self.vsb_tree_levels.set)
        self.my_list = os.listdir(
            self.install_dir / "Mods" / "Extracted" / "Data" / "Levels"
        )
        self.tree_levels.grid(row=0, column=0, rowspan=5, sticky="nswe")
        self.vsb_tree_levels.grid(row=0, column=0, rowspan=5, sticky="nse")

        self.mag = 50  # the size of each tiles in the grid; 50 is optimal
        self.rows = (
            15  # default values, could be set to none and still work I think lol
        )
        self.cols = (
            15  # default values, could be set to none and still work I think lol
        )

        self.canvas_grids = tk.Canvas(  # this is the main level editor grid
            self.tab2,
            bg="#292929",
        )
        self.canvas_grids.grid(row=0, column=1, rowspan=4, columnspan=2, sticky="nwse")

        self.canvas_grids.columnconfigure(0, weight=1)
        self.canvas_grids.rowconfigure(0, weight=1)

        self.scrollable_canvas_frame = tk.Frame(self.canvas_grids, bg="#343434")

        # offsets the screen so user can freely scroll around work area
        self.scrollable_canvas_frame.columnconfigure(
            0, minsize=int(int(self.screen_width) / 2)
        )
        self.scrollable_canvas_frame.columnconfigure(1, weight=1)
        self.scrollable_canvas_frame.columnconfigure(2, minsize=50)
        self.scrollable_canvas_frame.columnconfigure(
            4, minsize=int(int(self.screen_width) / 2)
        )
        self.scrollable_canvas_frame.rowconfigure(
            0, minsize=int(int(self.screen_height) / 2)
        )
        self.scrollable_canvas_frame.rowconfigure(1, weight=1)
        self.scrollable_canvas_frame.rowconfigure(2, minsize=100)
        self.scrollable_canvas_frame.rowconfigure(2, minsize=100)
        self.scrollable_canvas_frame.rowconfigure(
            4, minsize=int(int(self.screen_height) / 2)
        )

        self.scrollable_canvas_frame.grid(row=0, column=0, sticky="nwes")

        self.foreground_label = tk.Label(
            self.scrollable_canvas_frame,
            text="Foreground Area",
            fg="white",
            bg="#343434",
        )
        self.foreground_label.grid(row=2, column=1, sticky="nwse")
        self.foreground_label.grid_remove()

        self.background_label = tk.Label(
            self.scrollable_canvas_frame,
            text="Background Area",
            fg="white",
            bg="#343434",
        )
        self.background_label.grid(row=2, column=3, sticky="nwse")
        self.background_label.grid_remove()

        self.vbar = ttk.Scrollbar(
            self.tab2, orient="vertical", command=self.canvas_grids.yview
        )
        self.vbar.grid(row=0, column=2, rowspan=4, columnspan=1, sticky="nse")
        self.hbar = ttk.Scrollbar(
            self.tab2, orient="horizontal", command=self.canvas_grids.xview
        )
        self.hbar.grid(row=0, column=1, rowspan=4, columnspan=2, sticky="wes")

        self.canvas_grids.config(
            xscrollcommand=self.hbar.set, yscrollcommand=self.vbar.set
        )
        x0 = self.canvas_grids.winfo_screenwidth() / 2
        y0 = self.canvas_grids.winfo_screenheight() / 2
        self.canvas_grids.create_window(
            (x0, y0), window=self.scrollable_canvas_frame, anchor="center"
        )
        self.scrollable_canvas_frame.bind(
            "<Configure>",
            lambda e: self.canvas_grids.configure(
                scrollregion=self.canvas_grids.bbox("all")
            ),
        )

        self.canvas = tk.Canvas(  # this is the main level editor grid
            self.scrollable_canvas_frame,
            bg="#343434",
        )
        self.canvas.grid(row=1, column=1)
        self.canvas.grid_remove()
        self.canvas_dual = tk.Canvas(  # this is for dual level, it shows the back area
            self.scrollable_canvas_frame,
            width=0,
            bg="yellow",
        )
        self.canvas_dual.grid(row=1, column=3, padx=(0, 50))
        self.canvas_dual.grid_remove()  # hides it for now
        self.dual_mode = False

        self.tile_pallete = ScrollableFrame(  # the tile palletes are loaded into here as buttons with their image as a tile and txt as their value to grab when needed
            self.tab2, text="Tile Pallete", width=50
        )
        self.tile_pallete.grid(row=2, column=3, columnspan=4, rowspan=3, sticky="swne")
        self.tile_pallete.scrollable_frame["width"] = 50

        self.tile_label = tk.Label(
            self.tab2,
            text="Primary Tile:",
        )  # shows selected tile. Important because this is used for more than just user convenience; we can grab the currently used tile here
        self.tile_label.grid(row=0, column=4, columnspan=1, sticky="we")
        self.tile_label_secondary = tk.Label(
            self.tab2,
            text="Secondary Tile:",
        )  # shows selected tile. Important because this is used for more than just user convenience; we can grab the currently used tile here
        self.tile_label_secondary.grid(row=1, column=4, columnspan=1, sticky="we")

        self.button_tilecode_del = tk.Button(
            self.tab2,
            text="Del",
            bg="red",
            fg="white",
            width=10,
            command=self.del_tilecode,
        )
        self.button_tilecode_del.grid(row=0, column=3, sticky="e")
        self.button_tilecode_del["state"] = tk.DISABLED

        self.button_tilecode_del_secondary = tk.Button(
            self.tab2,
            text="Del",
            bg="red",
            fg="white",
            width=10,
            command=self.del_tilecode_secondary,
        )
        self.button_tilecode_del_secondary.grid(row=1, column=3, sticky="e")
        self.button_tilecode_del_secondary["state"] = tk.DISABLED

        self.img_sel = ImageTk.PhotoImage(
            Image.open(BASE_DIR / "static/images/tilecodetextures.png")
        )
        self.panel_sel = tk.Label(
            self.tab2, image=self.img_sel, width=50
        )  # shows selected tile image
        self.panel_sel.grid(row=0, column=5)
        self.panel_sel_secondary = tk.Label(
            self.tab2, image=self.img_sel, width=50
        )  # shows selected tile image
        self.panel_sel_secondary.grid(row=1, column=5)

        self.combobox = ttk.Combobox(self.tab2, height=20)
        self.combobox.grid(row=4, column=3, columnspan=1, sticky="nswe")
        self.combobox["state"] = tk.DISABLED
        self.combobox_alt = ttk.Combobox(self.tab2, height=40)
        self.combobox_alt.grid(row=4, column=4, columnspan=1, sticky="nswe")
        self.combobox_alt.grid_remove()
        self.combobox_alt["state"] = tk.DISABLED

        self.scale = tk.Scale(
            self.tab2, from_=0, to=100, orient=tk.HORIZONTAL, command=self.update_value
        )  # scale for the percent of a selected tile
        self.scale.grid(row=3, column=3, columnspan=2, sticky="we")
        self.scale.set(100)
        self.scale["state"] = tk.DISABLED

        self.button_tilecode_add = tk.Button(
            self.tab2,
            text="Add TileCode",
            bg="yellow",
            command=lambda: self.add_tilecode(
                str(self.combobox.get()), str(self.scale.get()), self.combobox_alt.get()
            ),
        )
        self.button_tilecode_add.grid(
            row=3, column=5, rowspan=2, columnspan=2, sticky="nswe"
        )

        self.tags = tk.Text(
            self.tab2, height=2, width=30, bg="black", fg="white"
        )  # text box for tags each room has like \!dual for example
        self.tags.grid(row=4, column=1, columnspan=2, sticky="we")

        # the tilecodes are in the same order as the tiles in the image(50x50, left to right)
        self.texture_images = []
        file_tile_codes = BASE_DIR / "tilecodes.txt"
        tile_lines = file_tile_codes.read_text(encoding="utf8").splitlines()
        count = 0
        count_total = 0
        x = 99
        y = 0
        # color_base = int(random.random())
        self.uni_tile_code_list = []
        self.tile_pallete_ref = []
        base_im = Image.open(BASE_DIR / "static/images/tilecodetextures.png")
        for line in tile_lines:
            line = line.strip()
            self.uni_tile_code_list.append(str(line))
            im = base_im.crop((count * 50, y * 50, 5000 - (x * 50), (1 + y) * 50))
            self.tile_texture = ImageTk.PhotoImage(im)
            self.texture_images.append(self.tile_texture)
            tile_ref = []
            tile_ref.append(line)
            tile_ref.append(self.texture_images[count_total])
            self.tile_pallete_ref.append(tile_ref)
            count = count + 1
            x -= 1
            if (
                count == 100
            ):  # theres a 100 tiles in each row on the image so this lets me know when to start grabbing from the next row
                y += 1
                x = 99
                count = 0
            count_total = count_total + 1
        self.panel_sel["image"] = self.texture_images[0]
        self.tile_label["text"] = "Primary Tile: " + r"\?empty a"
        self.panel_sel_secondary["image"] = self.texture_images[0]
        self.tile_label_secondary["text"] = "Secondary Tile: " + r"\?empty a"

        def canvas_click(event, canvas):  # when the level editor grid is clicked
            # Get rectangle diameters
            col_width = self.mag
            row_height = self.mag
            col = 0
            row = 0
            if canvas == self.canvas_dual:
                col = ((event.x + int(self.canvas["width"])) + col_width) // col_width
                row = event.y // row_height
            else:
                # Calculate column and row number
                col = event.x // col_width
                row = event.y // row_height
            # If the tile is not filled, create a rectangle
            if self.dual_mode:
                if int(col) == int((len(self.tiles[0]) - 1) / 2):
                    print("Middle of dual detected; not tile placed")
                    return

            canvas.delete(self.tiles[int(row)][int(col)])
            self.tiles[row][col] = canvas.create_image(
                int(col) * self.mag,
                int(row) * self.mag,
                image=self.panel_sel["image"],
                anchor="nw",
            )
            self.tiles_meta[row][col] = self.tile_label["text"].split(" ", 4)[3]
            print(
                str(self.tiles_meta[row][col])
                + " replaced with "
                + self.tile_label["text"].split(" ", 4)[3]
            )
            self.remember_changes()  # remember changes made

        def canvas_click_secondary(
            event, canvas
        ):  # when the level editor grid is clicked
            # Get rectangle diameters
            col_width = self.mag
            row_height = self.mag
            col = 0
            row = 0
            if canvas == self.canvas_dual:
                col = ((event.x + int(self.canvas["width"])) + col_width) // col_width
                row = event.y // row_height
            else:
                # Calculate column and row number
                col = event.x // col_width
                row = event.y // row_height
            # If the tile is not filled, create a rectangle
            if self.dual_mode:
                if int(col) == int((len(self.tiles[0]) - 1) / 2):
                    print("Middle of dual detected; not tile placed")
                    return

            canvas.delete(self.tiles[int(row)][int(col)])
            self.tiles[row][col] = canvas.create_image(
                int(col) * self.mag,
                int(row) * self.mag,
                image=self.panel_sel_secondary["image"],
                anchor="nw",
            )
            self.tiles_meta[row][col] = self.tile_label_secondary["text"].split(" ", 4)[
                3
            ]
            print(
                str(self.tiles_meta[row][col])
                + " replaced with "
                + self.tile_label["text"].split(" ", 4)[3]
            )
            self.remember_changes()  # remember changes made

        self.canvas.bind("<Button-1>", lambda event: canvas_click(event, self.canvas))
        self.canvas.bind(
            "<B1-Motion>", lambda event: canvas_click(event, self.canvas)
        )  # These second binds are so the user can hold down their mouse button when painting tiles
        self.canvas.bind(
            "<Button-3>", lambda event: canvas_click_secondary(event, self.canvas)
        )
        self.canvas.bind(
            "<B3-Motion>", lambda event: canvas_click_secondary(event, self.canvas)
        )  # These second binds are so the user can hold down their mouse button when painting tiles
        self.canvas_dual.bind(
            "<Button-1>", lambda event: canvas_click(event, self.canvas_dual)
        )
        self.canvas_dual.bind(
            "<B1-Motion>", lambda event: canvas_click(event, self.canvas_dual)
        )
        self.canvas_dual.bind(
            "<Button-3>", lambda event: canvas_click_secondary(event, self.canvas_dual)
        )
        self.canvas_dual.bind(
            "<B3-Motion>", lambda event: canvas_click_secondary(event, self.canvas_dual)
        )

        def tree_filesitemclick(_event):
            if self.save_needed and self.last_selected_file is not None:
                msg_box = tk.messagebox.askquestion(
                    "Continue?",
                    "You have unsaved changes to "
                    + str(self.tree_files.item(self.last_selected_file, option="text"))
                    + "\nContinue without saving?",
                    icon="warning",
                )
                if msg_box == "yes":
                    self.save_needed = False
                    print("Entered new files witout saving")
                else:
                    self.tree_files.selection_set(self.last_selected_file)
                    return
            item_text = ""
            self.canvas.delete("all")
            self.canvas_dual.delete("all")
            self.canvas.grid_remove()
            self.canvas_dual.grid_remove()
            self.foreground_label.grid_remove()
            self.background_label.grid_remove()
            for item in self.tree_files.selection():
                self.last_selected_file = item
                item_text = self.tree_files.item(item, "text")
                self.read_lvl_file(item_text)

        self.tree_files.bind("<ButtonRelease-1>", tree_filesitemclick)

    def tile_pick(
        self, event, button_row, button_col
    ):  # When a tile is selected from the tile pallete
        selected_tile = self.tile_pallete.scrollable_frame.grid_slaves(
            button_row, button_col
        )[0]
        self.panel_sel["image"] = selected_tile["image"]
        self.tile_label["text"] = "Primary Tile: " + selected_tile["text"]

    def tile_pick_secondary(
        self, event, button_row, button_col
    ):  # When a tile is selected from the tile pallete
        selected_tile = self.tile_pallete.scrollable_frame.grid_slaves(
            button_row, button_col
        )[0]
        self.panel_sel_secondary["image"] = selected_tile["image"]
        self.tile_label_secondary["text"] = "Secondary Tile: " + selected_tile["text"]

    def get_codes_left(self):
        codes = ""
        for code in self.usable_codes:
            codes += str(code)
        print(str(len(self.usable_codes)) + " codes left (" + codes + ")")

    def save_changes(self):
        if self.save_needed:
            file_content = ""
            skip_line = False
            for entry in self.tree.get_children():
                values = self.tree.item(entry)["values"]
                if str(values[0]) == "COMMENT":
                    if not skip_line:
                        file_content += str(values[2])
                    else:
                        skip_line = False
                    if (
                        str(values[0]) == "//"
                        or str(values[0]).startswith("// !")
                        or str(values[0]).startswith("// S")
                    ):
                        file_content += "\n"
                    if str(values[2]).startswith("//  TILE CODES"):
                        file_content += "// ------------------------------"
                        for tilecode in self.tile_pallete_ref_in_use:
                            file_content += "\n" + str(tilecode[0])
                        skip_line = True
                        file_content += "\n"
                else:
                    file_content += (
                        str(values[0])
                        + " "
                        + str(values[1])
                        + " "
                        + str(values[2])
                        + "\n"
                    )

            file_content += "\n"
            for room_parent in self.tree_levels.get_children():
                file_content += "\n////////////////////////////////////////////////////////////////////////////////\n"
                file_content += str(self.tree_levels.item(room_parent)["text"])
                file_content += "\n////////////////////////////////////////////////////////////////////////////////\n"
                for room in self.tree_levels.get_children(room_parent):
                    room_check = self.tree_levels.item(room)["text"]
                    if room_check == "room" or room_check == "room (edited)":
                        room_data = self.tree_levels.item(room, option="values")

                        list_to_str = "\n".join(
                            [str(save_row) for save_row in room_data]
                        )
                        file_content += "\n" + list_to_str + "\n"
            path = None
            if not self.extracts_mode:
                path = (
                    self.lvls_path
                    + "/"
                    + str(self.tree_files.item(self.last_selected_file, option="text"))
                )
            else:
                print("adding to overrides")
                path = (
                    self.install_dir
                    / "Mods"
                    / "Overrides"
                    / str(self.tree_files.item(self.last_selected_file, option="text"))
                )
            f = open(path, "w")
            f.write(file_content)
            f.close()
            self.save_needed = False

            print("Saved")
        else:
            print("No changes to save")

    def remember_changes(self):  # remembers changes made to rooms
        try:
            item_iid = self.tree_levels.selection()[0]
            parent_iid = self.tree_levels.parent(item_iid)  # gets selected room
            if parent_iid:
                self.canvas.delete("all")
                self.canvas_dual.delete("all")
                new_room_data = ""
                if len(self.tags.get(1.0, tk.END).split("!", 5)) <= 1:
                    new_room_data = ""
                else:
                    tag_count = -1
                    for tag in self.tags.get(1.0, tk.END).strip().split(" ", 6):
                        if str(tag).startswith(r"\!"):
                            if tag_count >= 0:
                                new_room_data += "\n" + str(tag)
                            else:
                                new_room_data += str(tag)
                        tag_count = tag_count + 1
                for row in self.tiles_meta:
                    if new_room_data != "":
                        new_room_data += "\n"
                    for block in row:
                        if str(block) == "None":
                            new_room_data += str(" ")
                        else:
                            new_room_data += str(block)
                room_save = []
                for line in new_room_data.split("\n", 100):
                    room_save.append(line)
                # Put it back in with the upated values
                edited = self.tree_levels.insert(
                    parent_iid,
                    self.tree_levels.index(item_iid),
                    text="room (edited)",
                    values=room_save,
                )
                # Remove it from the tree
                self.tree_levels.delete(item_iid)
                self.tree_levels.selection_set(edited)
                self.room_select(None)
                print("temp saved: \n" + new_room_data)
                print("Changes remembered!")
                self.save_needed = True
        except Exception as err:  # pylint: disable=broad-except
            print("No room to temp save " + str(err))

    def del_tilecode(self):
        msg_box = tk.messagebox.askquestion(
            "Delete Tilecode?",
            "Are you sure you want to delete this Tilecode?\nAll of its placements will be replaced with air",
            icon="warning",
        )
        if msg_box == "yes":
            tile_id = self.tile_label["text"].split(" ", 3)[2]
            tile_code = self.tile_label["text"].split(" ", 3)[3]
            if tile_id == r"\?empty":
                tkMessageBox.showinfo("Uh Oh!", "Can't delete empty!")
                return

            for room_parent in self.tree_levels.get_children():
                for room in self.tree_levels.get_children(room_parent):
                    room_data = []
                    room_name = self.tree_levels.item(room, option="text")
                    room_rows = self.tree_levels.item(room, option="values")
                    for row in room_rows:
                        new_row = ""
                        if not str(row).startswith(r"\!"):
                            for replace_code in row:
                                if replace_code == tile_code:
                                    replace_code = "0"
                                    new_row += "0"
                                else:
                                    new_row += str(replace_code)
                        else:
                            new_row = str(row)
                        room_data.append(new_row)
                    # Put it back in with the upated values
                    edited = self.tree_levels.insert(
                        room_parent,
                        self.tree_levels.index(room),
                        text=str(room_name),
                        values=room_data,
                    )
                    # Remove it from the tree
                    self.tree_levels.delete(room)
                    if room == self.last_selected_room:
                        self.tree_levels.selection_set(edited)
                        self.last_selected_room = edited
                        self.room_select(None)
            print("Replaced " + tile_id + " in all rooms with air/empty")

            self.usable_codes.append(str(tile_code))
            print(
                str(tile_code) + " is now available for use"
            )  # adds tilecode back to list to be reused
            ii = 0
            for id_ in self.tile_pallete_ref_in_use:
                if str(tile_id) == str(
                    id_[0].split(" ", 2)[0]
                ):  # removes tilecode from list in use
                    # self.usable_codes.pop(ii)
                    self.tile_pallete_ref_in_use.remove(id_)
                    print("Deleted " + str(tile_id))
                ii = ii + 1
            self.populate_tilecode_pallete()
            new_selection = self.tile_pallete_ref_in_use[0]
            if str(self.tile_label["text"]).split(" ", 3)[2] == tile_id:
                self.tile_label["text"] = (
                    "Primary Tile: "
                    + str(new_selection[0]).split(" ", 2)[0]
                    + " "
                    + str(new_selection[0]).split(" ", 2)[1]
                )
                self.panel_sel["image"] = new_selection[1]
            if str(self.tile_label_secondary["text"]).split(" ", 3)[2] == tile_id:
                self.tile_label_secondary["text"] = (
                    "Secondary Tile: "
                    + str(new_selection[0]).split(" ", 2)[0]
                    + " "
                    + str(new_selection[0]).split(" ", 2)[1]
                )
                self.panel_sel_secondary["image"] = new_selection[1]

            self.get_codes_left()
            self.save_needed = True
        else:
            return

    def del_tilecode_secondary(self):
        msg_box = tk.messagebox.askquestion(
            "Delete Tilecode?",
            "Are you sure you want to delete this Tilecode?\nAll of its placements will be replaced with air",
            icon="warning",
        )
        if msg_box == "yes":
            tile_id = self.tile_label_secondary["text"].split(" ", 3)[2]
            tile_code = self.tile_label_secondary["text"].split(" ", 3)[3]
            if tile_id == r"\?empty":
                tkMessageBox.showinfo("Uh Oh!", "Can't delete empty!")
                return

            for room_parent in self.tree_levels.get_children():
                for room in self.tree_levels.get_children(room_parent):
                    room_data = []
                    room_name = self.tree_levels.item(room, option="text")
                    room_rows = self.tree_levels.item(room, option="values")
                    for row in room_rows:
                        new_row = ""
                        if not str(row).startswith(r"\!"):
                            for replace_code in row:
                                if replace_code == tile_code:
                                    replace_code = "0"
                                    new_row += "0"
                                else:
                                    new_row += str(replace_code)
                        else:
                            new_row = str(row)
                        room_data.append(new_row)
                    # Put it back in with the upated values
                    edited = self.tree_levels.insert(
                        room_parent,
                        self.tree_levels.index(room),
                        text=str(room_name),
                        values=room_data,
                    )
                    # Remove it from the tree
                    self.tree_levels.delete(room)
                    if room == self.last_selected_room:
                        self.tree_levels.selection_set(edited)
                        self.last_selected_room = edited
                        self.room_select(None)
            print("Replaced " + tile_id + " in all rooms with air/empty")

            self.usable_codes.append(str(tile_code))
            print(
                str(tile_code) + " is now available for use"
            )  # adds tilecode back to list to be reused
            ii = 0
            for id_ in self.tile_pallete_ref_in_use:
                if str(tile_id) == str(
                    id_[0].split(" ", 2)[0]
                ):  # removes tilecode from list in use
                    # self.usable_codes.pop(ii)
                    self.tile_pallete_ref_in_use.remove(id_)
                    print("Deleted " + str(tile_id))
                ii = ii + 1
            self.populate_tilecode_pallete()
            new_selection = self.tile_pallete_ref_in_use[0]
            if str(self.tile_label["text"]).split(" ", 3)[2] == tile_id:
                self.tile_label["text"] = (
                    "Primary Tile: "
                    + str(new_selection[0]).split(" ", 2)[0]
                    + " "
                    + str(new_selection[0]).split(" ", 2)[1]
                )
                self.panel_sel["image"] = new_selection[1]
            if str(self.tile_label_secondary["text"]).split(" ", 3)[2] == tile_id:
                self.tile_label_secondary["text"] = (
                    "Secondary Tile: "
                    + str(new_selection[0]).split(" ", 2)[0]
                    + " "
                    + str(new_selection[0]).split(" ", 2)[1]
                )
                self.panel_sel_secondary["image"] = new_selection[1]

            self.get_codes_left()
            self.save_needed = True
        else:
            return

    def add_tilecode(self, tile, percent, alt_tile):
        usable_code = None

        tile = str(tile.split(" ", 3)[0])
        alt_tile = str(alt_tile.split(" ", 3)[0])  # isolates the id
        new_tile_code = r"\?" + tile

        tile_image = None
        tile_image_alt = None
        if any(
            tile + " " in self.e[0] for self.e in self.tile_pallete_ref
        ):  # compares tile id to tile ids in universal pallete list
            tile_image = self.e[1]

        if int(percent) < 100:
            new_tile_code += "%" + percent
            # Have to use a temporary directory due to TCL/Tkinter is trying to write
            # to a file name, not a file handle, and windows doesn't support sharing the
            # file between processes
            with tempfile.TemporaryDirectory() as tempdir:
                tempdir_path = Path(tempdir)
                temp1 = tempdir_path / "temp1"
                temp2 = tempdir_path / "temp2"
                tile_image._PhotoImage__photo.write(temp1, format="png")

                image1 = Image.open(
                    temp1,
                ).convert("RGBA")
                tile_text = percent + "%"
                if alt_tile != "empty":
                    if any(
                        alt_tile + " " in self.g[0] for self.g in self.tile_pallete_ref
                    ):  # compares tile id to tile ids in pallete list
                        tile_image_alt = self.g[1]
                    new_tile_code += "%" + alt_tile
                    tile_text += "/" + str(100 - int(percent)) + "%"

                    tile_image_alt._PhotoImage__photo.write(temp2)

                    image2 = Image.open(temp2).convert("RGBA")
                    image2.crop([25, 0, 50, 50]).save("temp2.png")
                    image2 = Image.open(temp2).convert("RGBA")

                    offset = (25, 0)
                    image1.paste(image2, offset)
                # make a blank image for the text, initialized to transparent text color
                txt = Image.new("RGBA", (50, 50), (255, 255, 255, 0))

                # get a drawing context
                d = ImageDraw.Draw(txt)

                # draw text, half opacity
                d.text((6, 34), tile_text, fill=(0, 0, 0, 255))
                d.text((4, 34), tile_text, fill=(0, 0, 0, 255))
                d.text((6, 36), tile_text, fill=(0, 0, 0, 255))
                d.text((4, 36), tile_text, fill=(0, 0, 0, 255))
                d.text((5, 35), tile_text, fill=(255, 255, 255, 255))

                out = Image.alpha_composite(image1, txt)
            tile_image = ImageTk.PhotoImage(out)

        if any(
            str(new_tile_code + " ") in str(self.g[0].split(" ", 3)[0]) + " "
            for self.g in self.tile_pallete_ref_in_use
        ):  # compares tile id to tile ids in pallete list
            tkMessageBox.showinfo("Uh Oh!", "You already have that!")
            return

        if len(self.usable_codes) > 0:
            usable_code = self.usable_codes[0]
            i = 0
            for ii in self.usable_codes:
                if ii == usable_code:
                    self.usable_codes.remove(ii)
                i = i + 1
        else:
            tkMessageBox.showinfo(
                "Uh Oh!", "You've reached the tilecode limit; delete some to add more"
            )
            return

        count_row = 0
        count_col = 0
        for i in self.tile_pallete_ref_in_use:
            if count_col == 7:
                count_col = -1
                count_row = count_row + 1
            count_col = count_col + 1

        ref_tile = []
        ref_tile.append(new_tile_code + " " + str(usable_code))
        ref_tile.append(tile_image)
        self.tile_pallete_ref_in_use.append(ref_tile)
        new_tile = tk.Button(
            self.tile_pallete.scrollable_frame,
            text=str(
                new_tile_code + " " + str(usable_code)
            ),  # keep seperate by space cause I use that for splitting
            width=40,
            height=40,
            image=tile_image,
        )
        new_tile.grid(row=count_row, column=count_col)
        new_tile.bind(
            "<Button-1>",
            lambda event, r=count_row, c=count_col: self.tile_pick(event, r, c),
        )
        new_tile.bind(
            "<Button-3>",
            lambda event, r=count_row, c=count_col: self.tile_pick_secondary(
                event, r, c
            ),
        )
        self.get_codes_left()
        self.save_needed = True

    def read_lvl_file(self, lvl):
        self.last_selected_room = None
        self.usable_codes_string = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ<>?,./;':][{}-=_+()0123456789!@#$%^&*`~"
        self.usable_codes = []
        for code in self.usable_codes_string:
            self.usable_codes.append(code)

        for (
            widget
        ) in (
            self.tile_pallete.scrollable_frame.winfo_children()
        ):  # removes any old tiles that might be there from the last file
            widget.destroy()

        for (
            i
        ) in (
            self.tree.get_children()
        ):  # removes any old rules that might be there from the last file
            self.tree.delete(i)

        if not self.extracts_mode:
            file1 = open(self.lvls_path + "/" + lvl)
        else:
            if (self.overrides_path / lvl).exists():
                print("Found this lvl in overrides; loading it instead")
                file1 = open(self.overrides_path / lvl)
            else:
                file1 = open(self.lvls_path / lvl)
        lines = file1.readlines()

        self.lvl_bg_path = self.textures_dir / "bg_cave.png"  # bg of main level grid
        if lvl == "abzu.lvl" or lvl.startswith("lake") or lvl.startswith("tide"):
            self.lvl_bg_path = self.textures_dir / "bg_tidepool.png"
        elif (
            lvl.startswith("babylon")
            or lvl.startswith("hallofu")
            or lvl.startswith("tiamat")
        ):
            self.lvl_bg_path = self.textures_dir / "bg_babylon.png"
        elif lvl.startswith("basecamp"):
            self.lvl_bg_path = self.textures_dir / "bg_cave.png"
        elif lvl.startswith("beehive"):
            self.lvl_bg_path = self.textures_dir / "bg_beehive.png"
        elif (
            lvl.startswith("blackmark")
            or lvl.startswith("jungle")
            or lvl.startswith("challenge_moon")
        ):
            self.lvl_bg_path = self.textures_dir / "bg_jungle.png"
        elif lvl.startswith("challenge_star"):
            self.lvl_bg_path = self.textures_dir / "bg_temple.png"
        elif (
            lvl.startswith("challenge_sun")
            or lvl.startswith("sunken")
            or lvl.startswith("hundun")
        ):
            self.lvl_bg_path = self.textures_dir / "bg_sunken.png"
        elif lvl.startswith("city"):
            self.lvl_bg_path = self.textures_dir / "bg_gold.png"
        elif lvl.startswith("egg"):
            self.lvl_bg_path = self.textures_dir / "bg_eggplant.png"
        elif lvl.startswith("ending_hard"):
            self.lvl_bg_path = self.textures_dir / "bg_cave.png"
        elif lvl.startswith("end"):
            self.lvl_bg_path = self.textures_dir / "bg_cave.png"
        elif lvl.startswith("ice"):
            self.lvl_bg_path = self.textures_dir / "bg_ice.png"
        elif lvl.startswith("olmec"):
            self.lvl_bg_path = self.textures_dir / "bg_stone.png"
        elif lvl.startswith("temple"):
            self.lvl_bg_path = self.textures_dir / "bg_temple.png"
        elif lvl.startswith("vlad"):
            self.lvl_bg_path = self.textures_dir / "bg_vlad.png"
        elif lvl.startswith("volcano"):
            self.lvl_bg_path = self.textures_dir / "bg_volcano.png"

        # Strips the newline character
        self.tree.delete(*self.tree.get_children())
        self.tree_levels.delete(*self.tree_levels.get_children())
        # pointer = 0
        # pointed = False

        # Enables widgets to use
        self.scale["state"] = tk.NORMAL
        self.combobox["state"] = tk.NORMAL
        self.combobox_alt["state"] = tk.NORMAL
        self.button_tilecode_del["state"] = tk.NORMAL
        self.button_tilecode_del_secondary["state"] = tk.NORMAL

        self.combobox_alt.grid_remove()
        self.scale.set(100)
        self.combobox.set(r"empty")
        self.combobox_alt.set(r"empty")

        self.tree_levels.bind("<ButtonRelease-1>", self.room_select)
        self.tile_pallete_ref_in_use = []
        self.node = self.tree_levels.insert("", "end", text="placeholder")

        def parse_lvl_file(lines, grab_parent_tilecode):
            load_mode = ""
            # blocks = []
            # cur_item = self.tree.focus()
            self.child = None
            self.rooms = []
            pointer = 0
            pointed = False
            # tile_count = 0
            # tl_col = 0
            # color_base = int(random.random())
            self.tile_convert = []
            line_count = 0
            line_total_count = len(lines)

            count = 0
            self.count_col = 0
            self.count_row = 0
            count_total = 0
            x = 99
            # y = 0

            for line in lines:
                line_raw = line
                line = " ".join(line.split())  # remove duplicate spaces
                line = line.strip()
                line_count = line_count + 1
                print("parsing " + line)
                if (
                    line == "// ------------------------------"
                    and not pointed
                    and pointer == 0
                    and load_mode != "Templates"
                    and line
                ):
                    pointer = pointer + 1
                    if not grab_parent_tilecode:
                        self.tree.insert(
                            "",
                            "end",
                            text="L1",
                            values=("COMMENT", "", str(line_raw)),
                        )
                elif pointer == 1 and not pointed and load_mode != "Templates" and line:
                    load_mode = "RoomChances"  # default because its usually at the top
                    pointed = True
                    if line == "// TILE CODES" and line:
                        load_mode = "TileCodes"
                    elif line == "// LEVEL CHANCES" and line:
                        load_mode = "LevelChances"
                    elif line == "// MONSTER CHANCES" and line:
                        load_mode = "MonsterChances"
                    elif (
                        line == "// TEMPLATES"
                        and line
                        or line == "// BASE CAMP (UNLOCKED SURFACE)"
                        and line
                    ):
                        load_mode = "Templates"
                    if not grab_parent_tilecode:
                        self.tree.insert(
                            "",
                            "end",
                            text="L1",
                            values=("COMMENT", "", str(line_raw)),
                        )
                elif (
                    line == "// ------------------------------"
                    and pointer == 1
                    # and load_mode != "Templates"
                    and line
                ):
                    pointer = 0
                    pointed = False
                    if not grab_parent_tilecode:
                        self.tree.insert(
                            "",
                            "end",
                            text="L1",
                            values=("COMMENT", "", str(line_raw)),
                        )
                elif (
                    load_mode == "RoomChances"
                    and not pointed
                    and pointer == 0
                    and line
                    and not grab_parent_tilecode
                ):
                    data = line.split(" ", 2)
                    comment = ""
                    value = ""
                    if len(data) > 1:
                        if str(data[0]) != r"\-size":
                            value = str(data[1])
                        else:
                            value = (
                                str(data[1]) + " " + str(str(data[2]).split("//", 1)[0])
                            )
                    comments = line.split("//", 1)
                    if len(comments) > 1:  # Makes sure a comment even exists
                        comment = "//   " + str(comments[1]).split("//", 2)[0]
                    self.tree.insert(
                        "",
                        "end",
                        text="L1",
                        values=(str(data[0]), str(value), str(comment)),
                    )
                elif load_mode == "TileCodes" and not pointed and pointer == 0 and line:
                    data = line.split(
                        " ", 2
                    )  # [0] = id [1] = tilecode [2] = comment if it exists
                    if not (
                        str(data[0]).startswith("//")
                    ):  # doesn't add tilecodes commented out
                        if not any(
                            " " + str(data[1]) in self.t[0]
                            for self.t in self.tile_pallete_ref_in_use
                        ):  # makes sure the code isn't in use already
                            tile_image = self.tile_pallete_ref[0][1]
                            if any(
                                str(data[0] + " ") in self.d
                                for self.d in self.uni_tile_code_list
                            ):  # compares tilecode id to its universal tilecode counterpart
                                if any(
                                    str(data[0] + " ") in self.b[0]
                                    for self.b in self.tile_pallete_ref
                                ):  # compares tile id to tile ids in pallete list
                                    tile_image = self.b[1]
                                ii = 0
                                for code in self.usable_codes:
                                    if str(data[1]) == str(
                                        code
                                    ):  # this removes the codes from the usable codes list cause it is now taken
                                        self.usable_codes.remove(code)
                                        print(str(code) + " is now taken!")
                                    ii = ii + 1
                                tile_ref = []
                                tile_ref.append(str(data[0]) + " " + str(data[1]))
                                tile_ref.append(tile_image)
                                self.tile_pallete_ref_in_use.append(tile_ref)
                                self.panel_sel["image"] = tile_image
                                self.tile_label["text"] = (
                                    "Primary Tile: " + str(data[0]) + " " + str(data[1])
                                )
                                self.panel_sel_secondary["image"] = tile_image
                                self.tile_label_secondary["text"] = (
                                    "Secondary Tile: "
                                    + str(data[0])
                                    + " "
                                    + str(data[1])
                                )
                            else:
                                try:
                                    if (
                                        len(str(data[0]).split("%", 2)) == 2
                                    ):  # tile with already existing percent
                                        if any(
                                            str(data[0]).split("%", 2)[0] + " "
                                            in self.e[0]
                                            for self.e in self.tile_pallete_ref
                                        ):  # compares tile id to tile ids in pallete list
                                            tile_image = self.e[1]
                                        tile_image._PhotoImage__photo.write("temp.png")
                                        self.ImageTk_image = Image.open(
                                            "temp.png"
                                        ).convert("RGBA")
                                        # make a blank image for the text, initialized to transparent text color
                                        txt = Image.new(
                                            "RGBA",
                                            self.ImageTk_image.size,
                                            (255, 255, 255, 0),
                                        )

                                        # get a drawing context
                                        d = ImageDraw.Draw(txt)

                                        # draw text, half opacity
                                        d.text(
                                            (4, 36),
                                            str(str(data[0]).split("%", 2)[1] + "%"),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (6, 36),
                                            str(str(data[0]).split("%", 2)[1] + "%"),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (4, 34),
                                            str(str(data[0]).split("%", 2)[1] + "%"),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (6, 34),
                                            str(str(data[0]).split("%", 2)[1] + "%"),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (5, 35),
                                            str(str(data[0]).split("%", 2)[1] + "%"),
                                            fill=(255, 255, 255, 255),
                                        )

                                        out = Image.alpha_composite(
                                            self.ImageTk_image, txt
                                        )
                                        self.done = ImageTk.PhotoImage(out)
                                        ii = 0
                                        for code in self.usable_codes:
                                            if str(data[1]) == str(
                                                code
                                            ):  # this removes the codes from the usable codes list cause it is now taken
                                                self.usable_codes.remove(code)
                                                print(str(code) + " is now taken!")
                                            ii = ii + 1
                                        tile_ref = []
                                        tile_ref.append(
                                            str(data[0]) + " " + str(data[1])
                                        )
                                        tile_ref.append(self.done)
                                        self.tile_pallete_ref_in_use.append(tile_ref)
                                        self.panel_sel["image"] = tile_image
                                        self.tile_label["text"] = (
                                            "Primary Tile: "
                                            + str(data[0])
                                            + " "
                                            + str(data[1])
                                        )
                                        self.panel_sel_secondary["image"] = tile_image
                                        self.tile_label_secondary["text"] = (
                                            "Secondary Tile: "
                                            + str(data[0])
                                            + " "
                                            + str(data[1])
                                        )
                                        if any(
                                            str(data[0]) + " " in self.e[0]
                                            for self.e in self.tile_pallete_ref_in_use
                                        ):  # compares tile id to tile ids in pallete list
                                            tile_image = self.e[1]
                                    elif (
                                        len(str(data[0]).split("%", 3)) == 3
                                    ):  # tile with already existing percent and alt tile
                                        tile_image = self.tile_pallete_ref[0][1]
                                        if any(
                                            str(data[0]).split("%", 2)[0] + " "
                                            in self.e[0]
                                            for self.e in self.tile_pallete_ref
                                        ):  # compares tile id to tile ids in pallete list
                                            tile_image = self.e[1]
                                        if any(
                                            r"\?" + str(data[0]).split("%", 2)[2] + " "
                                            in self.g[0]
                                            for self.g in self.tile_pallete_ref
                                        ):  # compares tile id to tile ids in pallete list
                                            tile_image_alt = self.g[1]

                                        tile_image._PhotoImage__photo.write("temp.png")
                                        tile_image_alt._PhotoImage__photo.write(
                                            "temp2.png"
                                        )

                                        self.ImageTk_image1 = Image.open(
                                            "temp.png"
                                        ).convert("RGBA")
                                        self.ImageTk_image2 = Image.open(
                                            "temp2.png"
                                        ).convert("RGBA")
                                        self.ImageTk_image2.crop([25, 0, 50, 50]).save(
                                            "temp2.png"
                                        )
                                        self.ImageTk_image2 = Image.open(
                                            "temp2.png"
                                        ).convert("RGBA")

                                        offset = (25, 0)
                                        self.ImageTk_image1.paste(
                                            self.ImageTk_image2, offset
                                        )
                                        # make a blank image for the text, initialized to transparent text color
                                        txt = Image.new(
                                            "RGBA", (50, 50), (255, 255, 255, 0)
                                        )

                                        # get a drawing context
                                        d = ImageDraw.Draw(txt)

                                        # draw text, half opacity
                                        d.text(
                                            (6, 34),
                                            str(
                                                str(data[0]).split("%", 2)[1]
                                                + "%/"
                                                + str(
                                                    100
                                                    - int(str(data[0]).split("%", 2)[1])
                                                )
                                                + "%"
                                            ),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (4, 34),
                                            str(
                                                str(data[0]).split("%", 2)[1]
                                                + "%/"
                                                + str(
                                                    100
                                                    - int(str(data[0]).split("%", 2)[1])
                                                )
                                                + "%"
                                            ),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (6, 36),
                                            str(
                                                str(data[0]).split("%", 2)[1]
                                                + "%/"
                                                + str(
                                                    100
                                                    - int(str(data[0]).split("%", 2)[1])
                                                )
                                                + "%"
                                            ),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (4, 36),
                                            str(
                                                str(data[0]).split("%", 2)[1]
                                                + "%/"
                                                + str(
                                                    100
                                                    - int(str(data[0]).split("%", 2)[1])
                                                )
                                                + "%"
                                            ),
                                            fill=(0, 0, 0, 255),
                                        )
                                        d.text(
                                            (5, 35),
                                            str(
                                                str(data[0]).split("%", 2)[1]
                                                + "%/"
                                                + str(
                                                    100
                                                    - int(str(data[0]).split("%", 2)[1])
                                                )
                                                + "%"
                                            ),
                                            fill=(255, 255, 255, 255),
                                        )

                                        out = Image.alpha_composite(
                                            self.ImageTk_image1, txt
                                        )
                                        self.done = ImageTk.PhotoImage(out)
                                        ii = 0
                                        for code in self.usable_codes:
                                            if str(data[1]) == str(
                                                code
                                            ):  # this removes the codes from the usable codes list cause it is now taken
                                                self.usable_codes.remove(code)
                                                print(str(code) + " is now taken!")
                                            ii = ii + 1
                                        tile_ref = []
                                        tile_ref.append(
                                            str(data[0]) + " " + str(data[1])
                                        )
                                        tile_ref.append(self.done)
                                        self.tile_pallete_ref_in_use.append(tile_ref)
                                        self.panel_sel["image"] = tile_image
                                        self.tile_label["text"] = (
                                            "Primary Tile: "
                                            + str(data[0])
                                            + " "
                                            + str(data[1])
                                        )
                                        self.panel_sel_secondary["image"] = tile_image
                                        self.tile_label_secondary["text"] = (
                                            "Secondary Tile: "
                                            + str(data[0])
                                            + " "
                                            + str(data[1])
                                        )
                                        if any(
                                            str(data[0]) + " " in self.r[0]
                                            for self.r in self.tile_pallete_ref_in_use
                                        ):  # compares tile id to tile ids in pallete list
                                            tile_image = self.r[1]
                                except Exception as err:  # pylint: disable=broad-except
                                    tkMessageBox.showinfo(
                                        "Uh Oh!",
                                        "skipped " + str(data[0]) + "\n" + str(err),
                                    )  # A tilecode id thats missing from the universal database and needs to be added
                                    # structure (id tilecode percent-without-percent-sign second-tile)

                            # new_image = ImageTk.PhotoImage(tile_image)
                            # new_image = new_image.text((10,10), "My Text", font=self.myFont, fill=(255,255,0))
                            if tile_image is not None:
                                count_row = 0
                                count_col = -1
                                for i in self.tile_pallete_ref_in_use:
                                    if count_col == 7:
                                        count_col = -1
                                        count_row = count_row + 1
                                    count_col = count_col + 1

                                new_tile = tk.Button(
                                    self.tile_pallete.scrollable_frame,
                                    text=str(
                                        data[0] + " " + data[1]
                                    ),  # keep seperate by space cause I use that for splitting
                                    width=40,
                                    height=40,
                                    image=tile_image,
                                )
                                new_tile.grid(row=count_row, column=count_col)
                                new_tile.bind(
                                    "<Button-1>",
                                    lambda event, r=count_row, c=count_col: self.tile_pick(
                                        event, r, c
                                    ),
                                )
                                new_tile.bind(
                                    "<Button-3>",
                                    lambda event, r=count_row, c=count_col: self.tile_pick_secondary(
                                        event, r, c
                                    ),
                                )
                                count = count + 1
                                self.count_col = self.count_col + 1
                                x -= 1
                                if self.count_col > 7:
                                    self.count_row = self.count_row + 1
                                    self.count_col = 0
                                count_total = count_total + 1
                elif (
                    load_mode == "LevelChances"
                    and not pointed
                    and pointer == 0
                    and line
                    and not grab_parent_tilecode
                ):
                    data = line.split(" ", 1)
                    comment = ""
                    value = ""
                    if len(data) > 1:
                        value = str(data[1]).split("//", 1)[0]
                    comments = line.split("//", 1)
                    if len(comments) > 1:  # Makes sure a comment even exists
                        comment = "//   " + comments[1]
                    self.tree.insert(
                        "",
                        "end",
                        text="L1",
                        values=(str(data[0]), str(value), str(comment)),
                    )
                elif (
                    load_mode == "MonsterChances"
                    and not pointed
                    and pointer == 0
                    and line
                    and not grab_parent_tilecode
                ):
                    data = line.split(" ", 1)
                    comment = ""
                    value = ""
                    if len(data) > 1:
                        value = str(data[1]).split("//", 1)[0]
                    comments = line.split("//", 1)
                    if len(comments) > 1:  # Makes sure a comment even exists
                        comment = "//   " + comments[1]
                    self.tree.insert(
                        "",
                        "end",
                        text="L1",
                        values=(str(data[0]), str(value), str(comment)),
                    )
                elif load_mode == "Templates" and not grab_parent_tilecode:
                    if line == "/" * 80 and not pointed and pointer == 0:
                        pointer = 1
                    elif not pointed and pointer == 1 and line.startswith(r"\."):
                        if (
                            self.tree_levels.item(self.tree_levels.get_children()[0])[
                                "text"
                            ]
                            == "placeholder"
                        ):
                            self.tree_levels.item(
                                self.tree_levels.get_children()[0],
                                values=line,
                                text=line,
                            )
                        else:
                            self.node = self.tree_levels.insert(
                                "", "end", values=line, text=line
                            )
                        pointed = True
                    elif line == "/" * 80 and pointed and pointer == 1:
                        pointed = False
                        pointer = 0
                    elif line.startswith("//") and not pointed and line:
                        self.tree.insert(
                            "", "end", text="L1", values=("COMMENT", "", line_raw)
                        )
                    elif not line and len(self.rooms) > 0:
                        self.new_room = self.tree_levels.insert(
                            self.node, "end", values=self.rooms, text="room"
                        )
                        self.rooms.clear()
                    elif line == "":
                        pointer = 0
                    elif line_count == line_total_count and len(self.rooms) > 0:
                        if line:
                            self.rooms.append(str(line))
                        self.new_room = self.tree_levels.insert(
                            self.node, "end", values=self.rooms, text="room"
                        )
                        self.rooms.clear()
                    elif line:
                        self.rooms.append(str(line))
                        print(str(line) + " parsed")

            file_id = self.tree_files.selection()[0]
            if str(self.tree_files.item(file_id, option="text")).startswith(
                "generic"
            ):  # adds tilecodes to generic that it relies on yet doesn't provide
                generic_needs = [
                    ["4", r"\?push_block"],
                    ["t", r"\?treasure"],
                    ["1", r"\?floor"],
                    ["6", r"\?chunk_air"],
                    ["=", r"\?minewood_floor"],
                ]
                for need in generic_needs:
                    for code in self.usable_codes:
                        if str(code) == need[0] and not any(
                            need[0] in str(code_in_use[0].split(" ", 3)[1])
                            for code_in_use in self.tile_pallete_ref_in_use
                        ):
                            if any(
                                need[1]
                                in str(str(self.universal_code[0]).split(" ", 3)[0])
                                for self.universal_code in self.tile_pallete_ref
                            ):
                                tile_ref = []
                                tile_ref.append(
                                    str(str(self.universal_code[0]).split(" ", 3)[0])
                                    + " "
                                    + need[0]
                                )
                                tile_ref.append(self.universal_code[1])
                                self.tile_pallete_ref_in_use.append(tile_ref)
                                print("Gave generic needed tilecode: " + need[1])
                self.populate_tilecode_pallete()

        parse_lvl_file(lines, False)  # parses to levels contents
        file_id = self.tree_files.selection()[0]
        lines_again = ""
        if str(self.tree_files.item(file_id, option="text")).startswith(
            "blackmark"
        ):  # goes back to grab levels parent data used in game
            if not self.extracts_mode:
                if os.path.isdir(self.lvls_path + "/" + "junglearea.lvl"):
                    lines_again = open(self.lvls_path + "/" + "junglearea.lvl")
                else:
                    print(
                        "local dependancy lvl not found, attempting load from extracts"
                    )
                    lines_again = open(self.extracts_path / "junglearea.lvl")
            else:
                if os.path.isdir(self.overrides_path / "junglearea.lvl"):
                    lines_again = open(self.overrides_path / "junglearea.lvl")
                else:
                    lines_again = open(self.lvls_path / "junglearea.lvl")
            parse_lvl_file(lines_again.readlines(), True)
        elif str(self.tree_files.item(file_id, option="text")).startswith(
            "vlads"
        ) or str(self.tree_files.item(file_id, option="text")).startswith(
            "challenge_moon"
        ):
            if not self.extracts_mode:
                if os.path.isdir(self.lvls_path + "/" + "volcanoarea.lvl"):
                    lines_again = open(self.lvls_path + "/" + "volcanoarea.lvl")
                else:
                    print(
                        "local dependancy lvl not found, attempting load from extracts"
                    )
                    lines_again = open(self.extracts_path / "volcanoarea.lvl")
            else:
                if os.path.isdir(self.overrides_path / "volcanoarea.lvl"):
                    lines_again = open(self.overrides_path / "volcanoarea.lvl")
                else:
                    lines_again = open(self.lvls_path / "volcanoarea.lvl")
            parse_lvl_file(lines_again.readlines(), True)
        elif (
            str(self.tree_files.item(file_id, option="text")).startswith("hallofush")
            or str(self.tree_files.item(file_id, option="text")).startswith(
                "babylonarea_1"
            )
            or str(self.tree_files.item(file_id, option="text")).startswith("palace")
        ):
            if not self.extracts_mode:
                if os.path.isdir(self.lvls_path + "/" + "babylonarea.lvl"):
                    lines_again = open(self.lvls_path + "/" + "babylonarea.lvl")
                else:
                    print(
                        "local dependancy lvl not found, attempting load from extracts"
                    )
                    lines_again = open(self.extracts_path / "babylonarea.lvl")
            else:
                if os.path.isdir(self.overrides_path / "babylonarea.lvl"):
                    lines_again = open(self.overrides_path / "babylonarea.lvl")
                else:
                    lines_again = open(self.lvls_path / "babylonarea.lvl")
            parse_lvl_file(lines_again.readlines(), True)
        elif str(self.tree_files.item(file_id, option="text")).startswith(
            "lake"
        ) or str(self.tree_files.item(file_id, option="text")).startswith(
            "challenge_star"
        ):
            if not self.extracts_mode:
                if os.path.isdir(self.lvls_path + "/" + "tidepoolarea.lvl"):
                    lines_again = open(self.lvls_path + "/" + "tidepoolarea.lvl")
                else:
                    print(
                        "local dependancy lvl not found, attempting load from extracts"
                    )
                    lines_again = open(self.extracts_path / "tidepoolarea.lvl")
            else:
                if os.path.isdir(self.overrides_path / "tidepoolarea.lvl"):
                    lines_again = open(self.overrides_path / "tidepoolarea.lvl")
                else:
                    lines_again = open(self.lvls_path / "tidepoolarea.lvl")
            parse_lvl_file(lines_again.readlines(), True)
        elif str(self.tree_files.item(file_id, option="text")).startswith("basecamp"):
            if not self.extracts_mode:
                if os.path.isdir(self.lvls_path + "/" + "basecamp.lvl"):
                    lines_again = open(self.lvls_path + "/" + "basecamp.lvl")
                else:
                    print(
                        "local dependancy lvl not found, attempting load from extracts"
                    )
                    lines_again = open(self.extracts_path / "basecamp.lvl")
            else:
                if os.path.isdir(self.overrides_path / "basecamp.lvl"):
                    lines_again = open(self.overrides_path / "basecamp.lvl")
                else:
                    lines_again = open(self.lvls_path / "basecamp.lvl")
            parse_lvl_file(lines_again.readlines(), True)
        elif str(self.tree_files.item(file_id, option="text")).startswith(
            "challenge_sun"
        ):
            if not self.extracts_mode:
                if os.path.isdir(self.lvls_path + "/" + "sunkencityarea.lvl"):
                    lines_again = open(self.lvls_path + "/" + "sunkencityarea.lvl")
                else:
                    print(
                        "local dependancy lvl not found, attempting load from extracts"
                    )
                    lines_again = open(self.extracts_path / "sunkencityarea.lvl")
            else:
                if os.path.isdir(self.overrides_path / "sunkencityarea.lvl"):
                    lines_again = open(self.overrides_path / "sunkencityarea.lvl")
                else:
                    lines_again = open(self.lvls_path / "sunkencityarea.lvl")
            parse_lvl_file(lines_again.readlines(), True)
        if not str(self.tree_files.item(file_id, option="text")).startswith(
            "basecamp"
        ) and not str(self.tree_files.item(file_id, option="text")).startswith(
            "generic"
        ):
            if not self.extracts_mode:
                if os.path.isdir(self.lvls_path + "/" + "generic.lvl"):
                    lines_again = open(self.lvls_path + "/" + "generic.lvl")
                else:
                    print(
                        "local dependancy lvl not found, attempting load from extracts"
                    )
                    lines_again = open(self.extracts_path / "generic.lvl")
            else:
                if os.path.isdir(self.overrides_path / "generic.lvl"):
                    lines_again = open(self.overrides_path / "generic.lvl")
                else:
                    lines_again = open(self.lvls_path / "generic.lvl")
            parse_lvl_file(
                lines_again.readlines(), True
            )  # finishes by grabbing parent generics tildecode data

        combo_tile_ids = []
        for tile_info in self.uni_tile_code_list:
            combo_tile_ids.append(str(tile_info).split(" ", 2)[0].replace(r"\?", ""))

        self.combobox["values"] = sorted(combo_tile_ids, key=str.lower)
        self.combobox_alt["values"] = sorted(combo_tile_ids, key=str.lower)
        # if self.tree.item(0, option="text")=="placeholder":
        # tree.delete(0)

    def update_value(self, _event):
        if int(self.scale.get()) == 100:
            self.combobox_alt.grid_remove()
            self.combobox.grid(columnspan=2)
        else:
            self.combobox.grid(columnspan=1)
            self.combobox_alt.grid()

    def _draw_grid(self, cols, rows, canvas, dual):
        # resizes canvas for grids
        canvas["width"] = (self.mag * cols) - 3
        canvas["height"] = (self.mag * rows) - 3

        if not dual:  # applies normal bg image settings to main grid
            self.cur_lvl_bg_path = (
                self.lvl_bg_path
            )  # store as a temp dif variable so it can switch back to the normal bg when needed

            file_id = self.tree_files.selection()[0]
            room_item = self.tree_levels.selection()[0]
            room_id = self.tree_levels.parent(
                room_item
            )  # checks which room is being opened to see if a special bg is needed
            factor = 1.0  # keeps image the same
            if self.lvl_bg_path == self.textures_dir / "bg_ice.png" and str(
                self.tree_levels.item(room_id, option="text")
            ).startswith(
                r"\.setroom1"
            ):  # mothership rooms are setroom10-1 to setroom13-2
                self.cur_lvl_bg_path = self.textures_dir / "bg_mothership.png"
            elif str(self.tree_files.item(file_id, option="text")).startswith(
                "blackmark"
            ):
                factor = 2.5  # brightens the image for black market
            elif (
                str(self.tree_files.item(file_id, option="text")).startswith("generic")
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "cosmic"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith("duat")
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "palace"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "ending_hard"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "challenge_m"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "challenge_st"
                )
            ):
                factor = 0  # darkens the image for cosmic ocean and duat and others

            image = Image.open(self.cur_lvl_bg_path)
            image = image.resize(
                (int(canvas["width"]), int(canvas["height"])), Image.BILINEAR
            )  ## The (250, 250) is (height, width)
            enhancer = ImageEnhance.Brightness(image)

            self.im_output = enhancer.enhance(factor)

            self.lvl_bg = ImageTk.PhotoImage(self.im_output)
            canvas.create_image(0, 0, image=self.lvl_bg, anchor="nw")
        else:  # applies special image settings if working with dual grid
            self.lvl_bgbg_path = (
                self.lvl_bg_path
            )  # Creates seperate image path variable for bgbg image

            file_id = self.tree_files.selection()[0]
            room_item = self.tree_levels.selection()[0]
            room_id = self.tree_levels.parent(
                room_item
            )  # checks which room is being opened to see if a special bg is needed
            factor = 0.6  # darkens the image
            if self.lvl_bg_path == self.textures_dir / "bg_ice.png":
                if str(self.tree_levels.item(room_id, option="text")).startswith(
                    r"\.mothership"
                ):
                    self.lvl_bgbg_path = self.textures_dir / "bg_mothership.png"
                    factor = 1.0  # keeps image the same
                else:
                    factor = 2.5  # brightens the image for ices caves
            elif str(self.tree_files.item(file_id, option="text")).startswith(
                "blackmark"
            ):
                factor = 2.5  # brightens the image for black market
            elif (
                str(self.tree_files.item(file_id, option="text")).startswith("generic")
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "cosmic"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith("duat")
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "palace"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "ending_hard"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "challenge_m"
                )
                or str(self.tree_files.item(file_id, option="text")).startswith(
                    "challenge_st"
                )
            ):
                factor = 0  # darkens the image for cosmic ocean and duat and others

            image_dual = Image.open(self.lvl_bgbg_path)
            image_dual = image_dual.resize(
                (int(canvas["width"]), int(canvas["height"])), Image.BILINEAR
            )  ## The (250, 250) is (height, width)
            enhancer = ImageEnhance.Brightness(image_dual)

            self.im_output_dual = enhancer.enhance(factor)

            self.lvl_bgbg = ImageTk.PhotoImage(self.im_output_dual)
            canvas.create_image(0, 0, image=self.lvl_bgbg, anchor="nw")

        # finishes by drawing grid on top
        for i in range(0, cols + 2):
            canvas.create_line(
                (i) * self.mag,
                0,
                (i) * self.mag,
                (rows) * self.mag,
                fill="#F0F0F0",
            )
        for i in range(0, rows):
            canvas.create_line(
                0,
                (i) * self.mag,
                self.mag * (cols + 2),
                (i) * self.mag,
                fill="#F0F0F0",
            )

    def room_select(self, _event):  # Loads room when click if not parent node
        self.dual_mode = False
        item_iid = self.tree_levels.selection()[0]
        parent_iid = self.tree_levels.parent(item_iid)
        if parent_iid:
            self.last_selected_room = item_iid
            self.canvas.delete("all")
            self.canvas_dual.delete("all")
            current_room = self.tree_levels.item(item_iid, option="values")
            current_room_tiles = []
            tags = ""
            self.tags.delete(1.0, "end")
            for cr_line in current_room:
                if str(cr_line).startswith(r"\!"):
                    print("found tag " + str(cr_line))
                    tags += cr_line + " "
                else:
                    print("appending " + str(cr_line))
                    current_room_tiles.append(str(cr_line))
                    for char in str(cr_line):
                        if str(char) == " ":
                            self.dual_mode = True

            self.tags.insert(1.0, tags)

            self.rows = len(current_room_tiles)
            self.cols = len(str(current_room_tiles[0]))

            print(
                str(self.rows)
                + " "
                + str(self.cols)
                + "-------------------------------------------------"
            )

            # self.mag = self.canvas.winfo_height() / self.rows - 30
            if not self.dual_mode:
                self._draw_grid(
                    self.cols, self.rows, self.canvas, False
                )  # cols rows canvas dual(True/False)
                self.canvas_dual["width"] = 0
                self.canvas_dual["height"] = 0
                self.canvas.grid()
                self.canvas_dual.grid_remove()  # hides it for now
                self.foreground_label.grid_remove()
                self.background_label.grid_remove()
            else:
                self.canvas.grid()
                self.canvas_dual.grid()  # brings it back
                self._draw_grid(
                    int((self.cols - 1) / 2), self.rows, self.canvas, False
                )  # cols rows canvas dual(True/False)
                self._draw_grid(
                    int((self.cols - 1) / 2), self.rows, self.canvas_dual, True
                )
                self.foreground_label.grid()
                self.background_label.grid()

            # Create a grid of None to store the references to the tiles
            self.tiles = [
                [None for _ in range(self.cols)] for _ in range(self.rows)
            ]  # tile image displays
            self.tiles_meta = [
                [None for _ in range(self.cols)] for _ in range(self.rows)
            ]  # meta for tile

            currow = -1
            curcol = 0
            for room_row in current_room_tiles:
                curcol = 0
                currow = currow + 1
                tile_image = None
                print(room_row)
                for block in str(room_row):
                    if str(block) != " ":
                        for pallete_block in self.tile_pallete_ref_in_use:
                            if any(
                                str(" " + block) in str(self.c[0])
                                for self.c in self.tile_pallete_ref_in_use
                            ):
                                tile_image = self.c[1]
                            else:
                                print(
                                    str(block) + " Not Found"
                                )  # There's a missing tile id somehow
                        if self.dual_mode and curcol > int((self.cols - 1) / 2):
                            x = int(curcol - ((self.cols - 1) / 2) - 1)
                            self.tiles[currow][curcol] = self.canvas_dual.create_image(
                                x * self.mag,
                                currow * self.mag,
                                image=tile_image,
                                anchor="nw",
                            )
                            coords = (
                                x * self.mag,
                                currow * self.mag,
                                x * self.mag + 50,
                                currow * self.mag + 50,
                            )
                            self.tiles_meta[currow][curcol] = block
                        else:
                            self.tiles[currow][curcol] = self.canvas.create_image(
                                curcol * self.mag,
                                currow * self.mag,
                                image=tile_image,
                                anchor="nw",
                            )
                            coords = (
                                curcol * self.mag,
                                currow * self.mag,
                                curcol * self.mag + 50,
                                currow * self.mag + 50,
                            )
                            self.tiles_meta[currow][curcol] = block
                        # print("loaded layer col " + str(curcol) + " " + str(self.c[0]) + " out of " + str(len(str(room_row))))

                    curcol = curcol + 1

    def on_double_click(self, tree_view):
        # First check if a blank space was selected
        entry_index = tree_view.focus()
        if entry_index == "":
            return

        # Set up window
        win = tk.Toplevel()
        win.title("Edit Entry")
        win.attributes("-toolwindow", True)
        self.center(win)

        ####
        # Set up the window's other attributes and geometry
        ####

        # Grab the entry's values
        for child in tree_view.get_children():
            if child == entry_index:
                values = tree_view.item(child)["values"]
                break

        col1_lbl = tk.Label(win, text="Entry: ")
        col1_ent = tk.Entry(win)
        col1_ent.insert(0, values[0])  # Default is column 1's current value
        col1_lbl.grid(row=0, column=0)
        col1_ent.grid(row=0, column=1)

        col2_lbl = tk.Label(win, text="Value: ")
        col2_ent = tk.Entry(win)
        col2_ent.insert(0, values[1])  # Default is column 2's current value
        col2_lbl.grid(row=0, column=2)
        col2_ent.grid(row=0, column=3)

        col3_lbl = tk.Label(win, text="Note: ")
        col3_ent = tk.Entry(win)
        col3_ent.insert(0, values[2])  # Default is column 3's current value
        col3_lbl.grid(row=0, column=4)
        col3_ent.grid(row=0, column=5)

        def update_then_destroy():
            if self.confirm_entry(
                tree_view, col1_ent.get(), col2_ent.get(), col3_ent.get()
            ):
                win.destroy()

        ok_button = tk.Button(win, text="Ok")
        ok_button.bind("<Button-1>", lambda e: update_then_destroy())
        ok_button.grid(row=1, column=2)

        cancel_button = tk.Button(win, text="Cancel")
        cancel_button.bind("<Button-1>", lambda c: win.destroy())
        cancel_button.grid(row=1, column=4)

    def confirm_entry(self, tree_view, entry1, entry2, entry3):
        ####
        # Whatever validation you need
        ####

        # Grab the current index in the tree
        current_index = tree_view.index(tree_view.focus())

        # Remove it from the tree
        self.delete_current_entry(tree_view)

        # Put it back in with the upated values
        tree_view.insert("", current_index, values=(entry1, entry2, entry3))
        self.save_needed = True

        return True

    def delete_current_entry(self, tree_view):
        curr = tree_view.focus()

        if curr == "":
            return

        tree_view.delete(curr)

    def center(self, toplevel):
        toplevel.update_idletasks()

        # Tkinter way to find the screen resolution
        # screen_width = toplevel.winfo_screenwidth()
        # screen_height = toplevel.winfo_screenheight()

        # find the screen resolution
        screen_width = int(self.screen_width)
        screen_height = int(self.screen_height)

        size = tuple(int(_) for _ in toplevel.geometry().split("+")[0].split("x"))
        x = screen_width / 2 - size[0] / 2
        y = screen_height / 2 - size[1] / 2

        toplevel.geometry("+%d+%d" % (x, y))

    def populate_tilecode_pallete(self):
        for (
            widget
        ) in (
            self.tile_pallete.scrollable_frame.winfo_children()
        ):  # resets tile pallete to add them all back without the deleted one
            widget.destroy()
        count_row = 0
        count_col = -1
        for tile_keep in self.tile_pallete_ref_in_use:
            if count_col == 7:
                count_col = -1
                count_row = count_row + 1
            count_col = count_col + 1

            new_tile = tk.Button(
                self.tile_pallete.scrollable_frame,
                text=str(tile_keep[0].split(" ", 2)[0])
                + " "
                + str(
                    tile_keep[0].split(" ", 2)[1]
                ),  # keep seperate by space cause I use that for splitting
                width=40,
                height=40,
                image=tile_keep[1],
            )
            new_tile.grid(row=count_row, column=count_col)
            new_tile.bind(
                "<Button-1>",
                lambda event, r=count_row, c=count_col: self.tile_pick(event, r, c),
            )
            new_tile.bind(
                "<Button-3>",
                lambda event, r=count_row, c=count_col: self.tile_pick_secondary(
                    event, r, c
                ),
            )

    def go_back(self):
        self.lvl_editor_start_canvas.grid()
        self.tab_control.grid_remove()
        self.tree_files.grid_remove()
        # Resets widgets
        self.scale["state"] = tk.DISABLED
        self.combobox["state"] = tk.DISABLED
        self.combobox_alt["state"] = tk.DISABLED
        self.button_tilecode_del["state"] = tk.DISABLED
        self.button_tilecode_del_secondary["state"] = tk.DISABLED
        self.canvas.delete("all")
        self.canvas_dual.delete("all")
        self.canvas.grid_remove()
        self.canvas_dual.grid_remove()
        self.foreground_label.grid_remove()
        self.background_label.grid_remove()
        self.button_back.grid_remove()
        self.button_save.grid_remove()
        self.vsb_tree_files.grid_remove()
        for (
            widget
        ) in (
            self.tile_pallete.scrollable_frame.winfo_children()
        ):  # removes any old tiles that might be there from the last file
            widget.destroy()



import os
import subprocess
import time
import urllib.request
from shutil import copyfile


def main():
    cwd = os.getcwd()

    copyfile(cwd + "/modlunky2.exe", cwd + "/modlunky2.exe.bak")  # Creates temp backup

    try:
        os.remove(cwd + "/modlunky2.exe")  # deletes current version
        print("Download latest release of Modlunky2..")
        # downlaods latest release on github
        url = "https://github.com/spelunky-fyi/modlunky2/releases/latest/download/modlunky2.exe"
        urllib.request.urlretrieve(url, cwd + "/modlunky2.exe")

        print("Download Complete!")
        # deletes backup once download is complete
        os.remove(cwd + "/modlunky2.exe.bak")

        # runs tool again and closes
        subprocess.call([cwd + "/modlunky2.exe"])

        # Wait for 5 seconds to give tool time to reopen
        time.sleep(5)
    except OSError:
        # restores backup if download failed for whatever reason
        copyfile(cwd + "/modlunky2.exe.bak", cwd + "modlunky2.exe")
        print("Download Failed.")


if __name__ == "__main__":
    main()



from pathlib import Path

from ..base_classes.base_sprite_loader import BaseSpriteLoader


class Big1(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbig01.png")
    _chunk_size = 128
    _chunk_map = {
        "cavemanboss": (0, 0, 2, 2),
        "giantspider": (0, 10, 2, 12),
        "queenbee": (0, 14, 2, 16),
    }


class Big2(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbig02.png")
    _chunk_size = 128
    _chunk_map = {
        "mummy": (0, 0, 2, 2),
        "anubis": (4, 8, 6, 11),
        "anubis2": (2, 8, 4, 10),
    }


class Big3(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbig03.png")
    _chunk_size = 128
    _chunk_map = {
        "lamassu": (0, 0, 2, 2),
        "yeti_king": (0, 4, 2, 6),
        "yeti_queen": (0, 10, 2, 12),
    }


class Big4(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbig04.png")
    _chunk_size = 128
    _chunk_map = {
        "crabman": (0, 0, 2, 2),
        "lavamander": (10, 4, 12, 6),
        "giantfly": (0, 12, 2, 14),
    }


class Big5(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbig05.png")
    _chunk_size = 128
    _chunk_map = {
        "ammit": (0, 4, 2, 5),
        "apep": (8, 0, 10, 2),
        "madametusk": (0, 8, 2, 10),
        "giant_frog": (0, 13, 3, 16),
        "minister": (0, 10, 1, 13),
    }


class Big6(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monstersbig06.png")
    _chunk_size = 128
    _chunk_map = {
        "kingu": (0, 0, 5, 6),
        "waddler": (0, 12, 2, 14),
        "humphead": (0, 14, 4, 16),
    }



from .base_sprite_loader import BaseSpriteLoader
from .base_deco_sheet import AbstractDecoSheet
from .base_biome import AbstractBiome, DEFAULT_BASE_PATH
from .base_floor_sheet import AbstractFloorSheet

__all__ = [
    "BaseSpriteLoader",
    "AbstractDecoSheet",
    "AbstractBiome",
    "AbstractFloorSheet",
    "DEFAULT_BASE_PATH",
]



"""
Biomes bring together the floor sheet, deco sheet, and background sprites for each
environment.
"""

from .base_classes import AbstractBiome
from .deco_sheet import *
from .floor_sheet import *

__all__ = [
    "CaveBiome",
    "VolcanaBiome",
    "JungleBiome",
    "TidePoolBiome",
    "TempleBiome",
    "IceCavesBiome",
    "BabylonBiome",
    "SunkenCityBiome",
    "EggplantBiome",
]


class CaveBiome(AbstractBiome):
    biome_name = "cave"
    display_name = "Dwelling"
    _floor_sheet_class = CaveFloorSheet
    _deco_sheet_class = CaveDecoSheet


class VolcanaBiome(AbstractBiome):
    biome_name = "volcano"
    display_name = "Volcana"
    _floor_sheet_class = VolcanaFloorSheet
    _deco_sheet_class = VolcanaDecoSheet


class JungleBiome(AbstractBiome):
    biome_name = "jungle"
    display_name = "Jungle"
    _floor_sheet_class = JungleFloorSheet
    _deco_sheet_class = JungleDecoSheet


class TidePoolBiome(AbstractBiome):
    biome_name = "tidepool"
    display_name = "Tide Pool"
    _floor_sheet_class = TidePoolFloorSheet
    _deco_sheet_class = TidePoolDecoSheet


class TempleBiome(AbstractBiome):
    biome_name = "temple"
    display_name = "Temple of Anubis"
    _floor_sheet_class = TempleFloorSheet
    _deco_sheet_class = TempleDecoSheet


class IceCavesBiome(AbstractBiome):
    biome_name = "ice"
    display_name = "Ice Caves"
    _floor_sheet_class = IceCavesFloorSheet
    _deco_sheet_class = TempleDecoSheet


class BabylonBiome(AbstractBiome):
    biome_name = "babylon"
    display_name = "Neo Babylon"
    _floor_sheet_class = BabylonFloorSheet
    _deco_sheet_class = BabylonDecoSheet


class SunkenCityBiome(AbstractBiome):
    biome_name = "sunken"
    display_name = "Sunken City"
    _floor_sheet_class = SunkenCityFloorSheet
    _deco_sheet_class = SunkenCityDecoSheet


class EggplantBiome(AbstractBiome):
    biome_name = "eggplant"
    display_name = "Eggplant World"
    _floor_sheet_class = EggplantFloorSheet
    _deco_sheet_class = EggplantDecoSheet



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Optional, Type

from PIL import Image

from .base_deco_sheet import AbstractDecoSheet
from .base_floor_sheet import AbstractFloorSheet

DEFAULT_BASE_PATH = Path(
    r"C:\Program Files (x86)\Steam\steamapps\common\Spelunky 2\Mods\Extracted"
)


class AbstractBiome(ABC):
    """Brings together the decoration sheet, floor sheet, and background for a _biome.
    Subclassed by overriding the abstracted methods with static class variables.

    Main apis are the `bg` method which returns the background image, and `get` which
    searches the two subclasses to see if it exists."""

    _image_cache: Dict[str, Image.Image]

    @property
    @abstractmethod
    def biome_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def display_name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def _floor_sheet_class(self) -> Type[AbstractFloorSheet]:
        raise NotImplementedError

    @property
    @abstractmethod
    def _deco_sheet_class(self) -> Type[AbstractDecoSheet]:
        raise NotImplementedError

    @property
    def bg(self) -> Image.Image:
        return self._bg

    def __init__(self, base_path: Path = DEFAULT_BASE_PATH):
        self.base_path = base_path
        self._floor_sheet = self._floor_sheet_class(base_path)
        self._deco_sheet = self._deco_sheet_class(base_path)
        self._bg = Image.open(base_path / f"Data/Textures/bg_{self.biome_name}.png")
        self._image_cache = {}
        self._sheet_map = self._make_sheet_map()

    # noinspection PyProtectedMember
    def _make_sheet_map(self) -> Dict:
        """
        Ensures the two subclasses don't overlap their key names, and builds a mapping
        so we know already where each key is and which class to ask for a key
        :return:
        """
        floor_keys = set(self._floor_sheet._chunk_map.keys())
        deco_keys = set(self._deco_sheet._chunk_map.keys())
        same_keys = floor_keys & deco_keys
        # Evals to false if nothing in the set
        if same_keys:
            raise KeyError(
                f"floor sheet and decoration sheet have colliding names: {same_keys}"
            )
        sheet_map = {
            k: self._floor_sheet.get for k in self._floor_sheet._chunk_map.keys()
        }
        for k in self._deco_sheet._chunk_map.keys():
            sheet_map[k] = self._deco_sheet.get
        return sheet_map

    def get(self, name: str) -> Optional[Image.Image]:
        # shortcut to check if we actually have this key available to us and bail out
        # early if we don't
        get_func = self._sheet_map.get(name)
        if get_func is None:
            return
        # this will either be the `get` method of the floor or deco sheet object
        return get_func(name)



from pathlib import Path

from .base_classes import BaseSpriteLoader


class JournalMonsterSheet(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/journal_entry_mons.png")
    _chunk_size = 160
    _chunk_map = {
        "journal_snake": (0, 0, 1, 1),
        "journal_spider": (1, 0, 2, 1),
        "journal_hang_spider": (2, 0, 3, 1),
        "journal_bat": (3, 0, 4, 1),
        "journal_cave_man": (4, 0, 5, 1),
        "journal_skeleton": (5, 0, 6, 1),
        "journal_scorpion": (6, 0, 7, 1),
        "journal_horned_lizard": (7, 0, 8, 1),
        "journal_man_trap": (8, 0, 9, 1),
        "journal_tiki_man": (9, 0, 10, 1),
        "journal_witch_doctor": (0, 1, 1, 2),
        "journal_mosquito": (1, 1, 2, 2),
        "journal_monkey": (2, 1, 3, 2),
        "journal_magmar": (3, 1, 4, 2),
        "journal_robot": (4, 1, 5, 2),
        "journal_fire_bug": (5, 1, 6, 2),
        "journal_vampire": (7, 1, 8, 2),
        "journal_vlad": (8, 1, 9, 2),
        "journal_croc_man": (9, 1, 10, 2),
        "journal_cobra": (0, 2, 1, 3),
        "journal_sorceress": (1, 2, 2, 3),
        "journal_cat_mummy": (2, 2, 3, 3),
        "journal_jiangshi": (3, 2, 4, 3),
        "journal_flying_fish": (4, 2, 5, 3),
        "journal_octopy": (5, 2, 6, 3),
        "journal_hermit_crab": (6, 2, 7, 3),
        "journal_ufo": (7, 2, 8, 3),
        "journal_alien": (8, 2, 9, 3),
        "journal_yeti": (9, 2, 10, 3),
        "journal_olmite": (0, 3, 1, 4),
        "journal_necromancer": (1, 3, 2, 4),
        "journal_bee": (2, 3, 3, 4),
        "journal_poochi": (4, 3, 5, 4),
        "journal_monty": (5, 3, 6, 4),
        "journal_golden_monkey": (6, 3, 7, 4),
        "journal_jiangshi_assassin": (7, 3, 8, 4),
        "journal_mole": (8, 3, 9, 4),
        "journal_percy": (9, 3, 10, 4),
        "journal_proto_shopkeeper": (0, 4, 1, 5),
        "journal_grub": (1, 4, 2, 5),
        "journal_frog": (2, 4, 3, 5),
        "journal_fire_frog": (3, 4, 4, 5),
        "journal_leprechain": (4, 4, 5, 5),
        "journal_egg_plup": (5, 4, 6, 5),
        "journal_ghist": (6, 4, 7, 5),
        "journal_tadpole": (7, 4, 8, 5),
        "journal_turkey": (8, 4, 9, 5),
        "journal_rockdog": (9, 4, 10, 5),
        "journal_axolotl": (0, 5, 1, 6),
        "journal_qilin": (1, 5, 2, 6),
    }



from modlunky2.sprites.base_classes.types import image_crop_tuple


def chunks_from_animation(
    base_name: str, start_chunk: image_crop_tuple, num_chunks: int
):
    chunk_width = start_chunk[2] - start_chunk[0]

    chunks = {}
    for i in range(0, num_chunks):
        coords = (
            start_chunk[0] + i * chunk_width,
            start_chunk[1],
            start_chunk[2] + i * chunk_width,
            start_chunk[3],
        )
        chunks["{}{}".format(base_name, i + 1)] = coords

    return chunks



from typing import Dict, Tuple, Union

_number_type = Union[float, int]
image_crop_tuple = Tuple[_number_type, _number_type, _number_type, _number_type]
chunk_map_type = Dict[str, image_crop_tuple]

image_crop_tuple_whole_number = Tuple[int, int, int, int]
chunk_map_type_whole_number = Dict[str, image_crop_tuple_whole_number]



import logging
import os
import queue
import tkinter as tk
from tkinter import ttk
from tkinter.scrolledtext import ScrolledText

logger = logging.getLogger("modlunky2")


class QueueHandler(logging.Handler):
    def __init__(self, log_queue):
        super().__init__()
        self.log_queue = log_queue

    def emit(self, record):
        self.log_queue.put(record)


# Adapted from https://beenje.github.io/blog/posts/logging-to-a-tkinter-scrolledtext-widget/
class ConsoleWindow(tk.Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Create a ScrolledText wdiget
        self.scrolled_text = ScrolledText(self, height=7, state="disabled")
        self.scrolled_text.pack(expand=True, fill="both")
        self.scrolled_text.configure(font="TkFixedFont")
        self.scrolled_text.tag_config("INFO", foreground="green")
        self.scrolled_text.tag_config("DEBUG", foreground="gray")
        self.scrolled_text.tag_config("WARNING", foreground="orange")
        self.scrolled_text.tag_config("ERROR", foreground="red")
        self.scrolled_text.tag_config("CRITICAL", foreground="red", underline=1)

        # Create a logging handler using a queue
        self.log_queue = queue.Queue()
        self.queue_handler = QueueHandler(self.log_queue)
        formatter = logging.Formatter("%(asctime)s: %(message)s")
        self.queue_handler.setFormatter(formatter)
        logger.addHandler(self.queue_handler)

        # Start polling messages from the queue
        self.after(100, self.poll_log_queue)

    def display(self, record):
        msg = self.queue_handler.format(record)
        self.scrolled_text.configure(state="normal")
        self.scrolled_text.insert(tk.END, msg + "\n", record.levelname)
        self.scrolled_text.configure(state="disabled")
        self.scrolled_text.yview(tk.END)

    def poll_log_queue(self):
        # Check every 100ms if there is a new message in the queue to display
        while True:
            try:
                record = self.log_queue.get(block=False)
            except queue.Empty:
                break
            else:
                self.display(record)
        self.after(100, self.poll_log_queue)

    def close(self):
        pass


class ToolTip:
    def __init__(self, widget, text):
        self.widget = widget
        self.text = text

        self.widget.bind('<Enter>', self.on_enter)
        self.widget.bind('<Leave>', self.on_leave)

        self.tooltip = None
        self.label = None

    def on_enter(self, event):
        self.tooltip = tk.Toplevel()
        self.tooltip.overrideredirect(True)
        self.tooltip.geometry(f'+{event.x_root+15}+{event.y_root+10}')

        self.label = tk.Label(self.tooltip,text=self.text)
        self.label.pack()

    def on_leave(self, _event):
        self.tooltip.destroy()


class Tab(ttk.Frame):
    """ Base class that all tabs should inherit from."""

    def on_load(self):
        """ Called whenever the tab is loaded."""


class ScrollableFrame(ttk.LabelFrame):
    def __init__(self, container, *args, **kwargs):
        super().__init__(container, *args, **kwargs)
        self.canvas = tk.Canvas(self)
        self.scrollbar = ttk.Scrollbar(
            self, orient="vertical", command=self.canvas.yview
        )
        self.scrollable_frame = tk.Frame(self.canvas)
        self.canvas.pack(side="left", fill="both", expand=True)

        self.scrollable_frame.bind(
            "<Configure>",
            lambda e: self.canvas.configure(scrollregion=self.canvas.bbox("all")),
        )

        self.canvas.create_window((0, 0), window=self.scrollable_frame, anchor="nw")

        self.canvas.configure(yscrollcommand=self.scrollbar.set)
        self.canvas.bind("<Enter>", self._bind_to_mousewheel)
        self.canvas.bind("<Leave>", self._unbind_from_mousewheel)

        self.canvas.pack(side="left", fill="both", expand=True)
        self.scrollbar.pack(side="right", fill="y")

    def _on_mousewheel(self, event):
        scroll_dir = None
        if event.num == 5 or event.delta == -120:
            scroll_dir = 1
        elif event.num == 4 or event.delta == 120:
            scroll_dir = -1

        if scroll_dir is None:
            return

        # If the scrollbar is max size don't bother scrolling
        if self.scrollbar.get() == (0.0, 1.0):
            return

        self.canvas.yview_scroll(scroll_dir, "units")

    def _bind_to_mousewheel(self, _event):
        if "nt" in os.name:
            self.canvas.bind_all("<MouseWheel>", self._on_mousewheel)
        else:
            self.canvas.bind_all("<Button-4>", self._on_mousewheel)
            self.canvas.bind_all("<Button-5>", self._on_mousewheel)

    def _unbind_from_mousewheel(self, _event):
        if "nt" in os.name:
            self.canvas.unbind_all("<MouseWheel>")
        else:
            self.canvas.unbind_all("<Button-4>")
            self.canvas.unbind_all("<Button-5>")


class ToggledFrame(tk.Frame):
    def __init__(self, parent, text, *args, **options):
        tk.Frame.__init__(self, parent, *args, **options)

        self.show = tk.IntVar()
        self.show.set(0)

        self.title_frame = ttk.Frame(self)
        self.title_frame.pack(fill="x", expand=1)

        ttk.Label(self.title_frame, text=text).pack(side="left", fill="x", expand=1)

        self.toggle_button = ttk.Checkbutton(
            self.title_frame,
            width=2,
            text="+",
            command=self.toggle,
            variable=self.show,
            style="Toolbutton",
        )
        self.toggle_button.pack(side="left")

        self.sub_frame = tk.Frame(self, relief="sunken", borderwidth=1)

    def toggle(self):
        if bool(self.show.get()):
            self.sub_frame.pack(fill="x", expand=1)
            self.toggle_button.configure(text="-")
        else:
            self.sub_frame.forget()
            self.toggle_button.configure(text="+")


class LevelsTree(ttk.Treeview):
    def __init__(self, parent, *args, **kwargs):
        ttk.Treeview.__init__(self, parent, *args, **kwargs)

        self.popup_menu = tk.Menu(self, tearoff=0)
        # self.popup_menu.add_command(label="Rename List Node",
        #            command=self.rename)
        # self.popup_menu.add_command(label="Delete Room", command=self.delete_selected)
        # self.popup_menu.add_command(label="Add Room", command=self.add_room)

        self.bind("<Button-3>", self.popup)  # Button-2 on Aqua

    def popup(self, event):
        try:
            self.popup_menu.tk_popup(event.x_root, event.y_root, 0)
        finally:
            self.popup_menu.grab_release()

    def rename(self):
        for i in self.selection()[::-1]:
            self.rename_dialog()

    # def delete_selected(self):
    #    item_iid = self.selection()[0]
    #    parent_iid = self.parent(item_iid)  # gets selected room
    #    if parent_iid:
    #        if (item_iid==LevelsTab.last_selected_room):
    #            LevelsTab.canvas.grid_remove()
    #            LevelsTab.canvas_dual.grid_remove()
    #            LevelsTab.foreground_label.grid_remove()
    #            LevelsTab.background_label.grid_remove()
    #        self.delete(item_iid)
    #
    # def add_room(self):
    #    item_iid = self.tree_levels.selection()[0]
    #    parent_iid = self.tree_levels.parent(item_iid)  # gets selected room
    #    if parent_iid:
    #        edited = self.insert(
    #            parent_iid,
    #            self.tree_levels.index(item_iid),
    #            text="room",
    #            values=room_save,
    #        )
    #    #self.selection_set(0, 'end')

    def rename_dialog(self):
        # First check if a blank space was selected
        entry_index = self.focus()
        if entry_index == "":
            return

        # Set up window
        win = tk.Toplevel()
        win.title("Edit Entry")
        win.attributes("-toolwindow", True)
        self.center(win)

        ####
        # Set up the window's other attributes and geometry
        ####

        # Grab the entry's values
        for child in self.get_children():
            if child == entry_index:
                values = self.item(child)["values"]
                break

        entry_name = str(values[0])
        entry_note = ""
        entry = str(values[0]).split("//", 2)
        if len(entry) > 1:
            entry_name = entry[0]
            entry_note = entry[1]

        col1_lbl = tk.Label(win, text="Entry: ")
        col1_ent = tk.Entry(win)
        col1_ent.insert(0, entry_name)  # Default is column 1's current value
        col1_lbl.grid(row=0, column=0)
        col1_ent.grid(row=0, column=1)

        col2_lbl = tk.Label(win, text="Note Name: ")
        col2_ent = tk.Entry(win)
        col2_ent.insert(0, entry_note)  # Default is column 2's current value
        col2_lbl.grid(row=0, column=2)
        col2_ent.grid(row=0, column=3)

        def update_then_destroy():
            if self.confirm_entry(self, col1_ent.get(), col2_ent.get()):
                win.destroy()

        ok_button = tk.Button(win, text="Ok")
        ok_button.bind("<Button-1>", lambda e: update_then_destroy())
        ok_button.grid(row=1, column=1)

        cancel_button = tk.Button(win, text="Cancel")
        cancel_button.bind("<Button-1>", lambda c: win.destroy())
        cancel_button.grid(row=1, column=3)

    def confirm_entry(self, entry1, entry2, entry3):
        ####
        # Whatever validation you need
        ####

        # Grab the current index in the tree
        current_index = self.index(self.focus())

        # Remove it from the tree
        self.delete(self.focus())

        # Put it back in with the upated values
        self.insert(
            "", current_index, values=(str("\\" + str(entry1) + " //" + str(entry2)))
        )
        self.save_needed = True

        return True

    def center(self, toplevel):
        toplevel.update_idletasks()

        # Tkinter way to find the screen resolution
        # screen_width = toplevel.winfo_screenwidth()
        # screen_height = toplevel.winfo_screenheight()

        # find the screen resolution
        screen_width = 1280
        screen_height = 720

        size = tuple(int(_) for _ in toplevel.geometry().split("+")[0].split("x"))
        x = screen_width / 2 - size[0] / 2
        y = screen_height / 2 - size[1] / 2

        toplevel.geometry("+%d+%d" % (x, y))


class RulesTree(ttk.Treeview):
    def __init__(self, parent, *args, **kwargs):
        ttk.Treeview.__init__(self, parent, *args, **kwargs)

        self.popup_menu = tk.Menu(self, tearoff=0)
        self.popup_menu.add_command(label="Add", command=self.add)
        self.popup_menu.add_command(label="Delete", command=self.delete_selected)

        self.bind("<Button-3>", self.popup)  # Button-2 on Aqua

    def popup(self, event):
        try:
            self.popup_menu.tk_popup(event.x_root, event.y_root, 0)
        finally:
            self.popup_menu.grab_release()

    def delete_selected(self):
        msg_box = tk.messagebox.askquestion(
            "Delete?",
            "Delete this rule?",
            icon="warning",
        )
        if msg_box == "yes":
            item_iid = self.selection()[0]
            self.delete(item_iid)

    def add(self):
        edited = self.insert(
            "",
            "end",
            values=["COMMENT", "VAL", "// COMMENT"],
        )
        # self.selection_set(0, 'end')



from dataclasses import dataclass
from collections import OrderedDict
from typing import ClassVar, Generic, Optional, TextIO, TypeVar

from .utils import DirectivePrefixes, split_comment, to_line

T = TypeVar("T")  # pylint: disable=invalid-name

VALID_LEVEL_SETTINGS = set(
    [
        "altar_room_chance",
        "back_room_chance",
        "back_room_hidden_door_cache_chance",
        "back_room_hidden_door_chance",
        "back_room_interconnection_chance",
        "background_chance",
        "flagged_liquid_rooms",
        "floor_bottom_spread_chance",
        "floor_side_spread_chance",
        "ground_background_chance",
        "idol_room_chance",
        "liquid_gravity",
        "machine_bigroom_chance",
        "machine_rewardroom_chance",
        "machine_tallroom_chance",
        "machine_wideroom_chance",
        "max_liquid_particles",
        "mount_chance",
        "size",
    ]
)

NAME_PADDING = max(map(len, VALID_LEVEL_SETTINGS)) + 4


class LevelSettings:
    def __init__(self):
        self._inner = OrderedDict()
        self.comment = None

    def get(self, name):
        LevelSetting.validate_name(name)
        return self._inner.get(name)

    def set_obj(self, obj: "LevelSetting"):
        obj.clean()
        obj.validate()
        self._inner[obj.name] = obj

    def write(self, handle: TextIO):
        if self.comment:
            handle.write(f"{self.comment}\n")
        for obj in self._inner.values():
            handle.write(obj.to_line())
        handle.write("\n")


@dataclass
class LevelSetting(Generic[T]):
    prefix: ClassVar[str] = DirectivePrefixes.LEVEL_SETTING.value
    name: str
    value: T
    comment: Optional[str]

    @classmethod
    def parse(cls, line: str) -> "LevelSetting":
        rest, comment = split_comment(line)
        directive, value = rest.split(None, 1)
        name = directive[2:]

        if not name:
            raise ValueError("Directive missing name.")

        level_setting = cls(name, value, comment)
        level_setting.clean()
        level_setting.validate()

        return level_setting

    @staticmethod
    def validate_name(name: str):
        if name not in VALID_LEVEL_SETTINGS:
            raise ValueError(f"Invalid name provided for LevelSetting: {name!r}")

    def validate_value(self):
        if self.name == "size":
            if not isinstance(self.value, tuple) or len(self.value) != 2:
                raise ValueError(
                    f"LevelSetting with name {self.name} expected a tuple of length 2: {self.value!r}"
                )
        elif self.name == "liquid_gravity":
            if not isinstance(self.value, float):
                raise ValueError(
                    f"LevelSetting with name {self.name} expected to be a float: {self.value!r}"
                )
        else:
            if not isinstance(self.value, int):
                raise ValueError(
                    f"LevelSetting with name {self.name} expected to be an int: {self.value!r}"
                )

    def validate(self):
        self.validate_name(self.name)
        self.validate_value()

    def clean_value(self):
        if not isinstance(self.value, str):
            return

        if self.name == "size":
            value = tuple(self.value.split())
            if len(value) != 2:
                raise ValueError("Directive `size` expects 2 values.")
        elif self.name == "liquid_gravity":
            value = float(self.value)
        else:
            value = int(self.value)

        self.value = value

    def clean(self):
        self.clean_value()

    def value_to_str(self) -> str:
        if isinstance(self.value, (int, float)):
            return f"{self.value}"
        return " ".join(self.value)

    def to_line(self) -> str:
        return to_line(
            self.prefix, self.name, NAME_PADDING, self.value_to_str(), 10, self.comment
        )

    def write(self, handle: TextIO):
        handle.write(self.to_line())



from pathlib import Path

from .base_classes import BaseSpriteLoader


class StickerSheet(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/journal_stickers.png")
    _chunk_size = 80
    _chunk_map = {
        "sticker_char_yellow": (0, 0, 1, 1),
        "sticker_char_magenta": (1, 0, 2, 1),
        "sticker_char_cyan": (2, 0, 3, 1),
        "sticker_char_black": (3, 0, 4, 1),
        "sticker_char_cinnabar": (4, 0, 5, 1),
        "sticker_char_green": (5, 0, 6, 1),
        "sticker_char_olive": (6, 0, 7, 1),
        "sticker_char_white": (7, 0, 8, 1),
        "sticker_char_cerulean": (8, 0, 9, 1),
        "sticker_char_blue": (9, 0, 10, 1),
        "sticker_char_lime": (0, 1, 1, 2),
        "sticker_char_lemon": (1, 1, 2, 2),
        "sticker_char_iris": (2, 1, 3, 2),
        "sticker_char_gold": (3, 1, 4, 2),
        "sticker_char_red": (4, 1, 5, 2),
        "sticker_char_pink": (5, 1, 6, 2),
        "sticker_char_violet": (6, 1, 7, 2),
        "sticker_char_gray": (7, 1, 8, 2),
        "sticker_char_khaki": (8, 1, 9, 2),
        "sticker_char_orange": (9, 1, 10, 2),
        "sticker_udjat_eye": (0, 2, 1, 3),
        "sticker_hedjet": (1, 2, 2, 3),
        "sticker_crown": (2, 2, 3, 3),
        "sticker_ankh": (3, 2, 4, 3),
        "sticker_kapala": (4, 2, 5, 3),
        "sticker_idol": (5, 2, 6, 3),
        "sticker_the_true_crown": (6, 2, 7, 3),
        "sticker_yellow_tile": (7, 2, 8, 3),
        "sticker_lock": (8, 2, 9, 3),
        "sticker_char_locked": (9, 2, 10, 3),
        "sticker_paste": (0, 3, 1, 4),
        "sticker_compass": (1, 3, 2, 4),
        "sticker_climbing_gloves": (2, 3, 3, 4),
        "sticker_pitchers_mitt": (3, 3, 4, 4),
        "sticker_spike_shoes": (4, 3, 5, 4),
        "sticker_spring_shoes": (5, 3, 6, 4),
        "sticker_parachute": (6, 3, 7, 4),
        "sticker_speckles": (7, 3, 8, 4),
        "sticker_alien_compass": (8, 3, 9, 4),
        "sticker_eggplant_crown": (9, 3, 10, 4),
        "sticker_cape": (0, 4, 1, 5),
        "sticker_jet_pack": (1, 4, 2, 5),
        "sticker_tele_pack": (2, 4, 3, 5),
        "sticker_hover_pack": (3, 4, 4, 5),
        "sticker_skeleton_key": (4, 4, 5, 5),
        "sticker_tablet_of_destiny": (5, 4, 6, 5),
        "sticker_power_pack": (6, 4, 7, 5),
        "sticker_vlad_cape": (7, 4, 8, 5),
        "sticker_50_percent": (8, 4, 9, 5),
        "sticker_full": (9, 4, 10, 5),
        "sticker_shopkeeper": (0, 5, 1, 6),
        "sticker_tun": (1, 5, 2, 6),
        "sticker_cave_man": (2, 5, 3, 6),
        "sticker_yang": (3, 5, 4, 6),
        "sticker_parsnip": (5, 5, 6, 6),
        "sticker_parmesan": (6, 5, 7, 6),
        "sticker_van_horsing": (7, 5, 8, 6),
        "sticker_sparrow": (8, 5, 9, 6),
        "sticker_vlad": (9, 5, 10, 6),
        "sticker_quill_back": (0, 6, 2, 8),
        "sticker_kingu": (2, 6, 4, 8),
        "sticker_anubis": (4, 6, 6, 8),
        "sticker_osiris": (6, 6, 8, 8),
        "sticker_lahamu": (8, 6, 10, 8),
        "sticker_olmec": (0, 8, 2, 10),
        "sticker_tiamat": (2, 8, 4, 10),
        "sticker_eggplant_king": (4, 8, 6, 10),
        "sticker_hundun": (6, 8, 8, 10),
        "sticker_blob_blue": (8, 8, 9, 9),
        "sticker_blob_green": (9, 8, 10, 9),
        "sticker_empty_1": (8, 9, 9, 10),
        "sticker_empty_2": (9, 9, 10, 10),
    }



import logging
import shutil
import threading
import tkinter as tk
import webbrowser
from pathlib import Path
from tkinter import ttk

from PIL import Image, ImageTk

from modlunky2.assets.assets import AssetStore
from modlunky2.assets.exc import MissingAsset
from modlunky2.assets.hashing import md5sum_path
from modlunky2.assets.patcher import Patcher
from modlunky2.constants import BASE_DIR
from modlunky2.ui.utils import is_patched, log_exception
from modlunky2.ui.widgets import ScrollableFrame, Tab, ToolTip

logger = logging.getLogger("modlunky2")

MODS = Path("Mods")


class PackTab(Tab):
    def __init__(self, tab_control, install_dir, *args, **kwargs):
        super().__init__(tab_control, *args, **kwargs)
        self.tab_control = tab_control
        self.install_dir = install_dir

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, minsize=60)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)


        self.frame = ScrollableFrame(self, text="Select mods to pack")
        self.frame.grid(row=0, column=0, columnspan=3, pady=5, padx=5, sticky="nswe")
        self.frame.rowconfigure(0, weight=1)
        self.frame.columnconfigure(0, weight=1)

        self.button_pack = ttk.Button(self, text="Pack", command=self.pack)
        self.button_pack.grid(row=1, column=0, pady=5, padx=5, sticky="nswe")
        ToolTip(self.button_pack, "Pack modded assets into the EXE.")

        self.button_restore = ttk.Button(self, text="Restore EXE", command=self.restore)
        self.button_restore.grid(row=1, column=1, pady=5, padx=5, sticky="nswe")
        ToolTip(self.button_restore, "Restore EXE to vanilla state from backup.")

        self.button_validate = ttk.Button(self, text="Validate Game Files", command=self.validate)
        self.button_validate.grid(row=1, column=2, pady=5, padx=5, sticky="nswe")
        ToolTip(self.button_validate, "Redownload vanilla EXE from steam.")

        default_icon_path = BASE_DIR / "static/images/noicon.png"
        self.default_icon = ImageTk.PhotoImage(Image.open(default_icon_path))

        self.checkbox_vars = []

    @staticmethod
    def validate():
        webbrowser.open_new_tab("steam://validate/418530")

    def restore(self):
        mods_dir = self.install_dir / MODS
        extract_dir = mods_dir / "Extracted"
        source_exe = extract_dir / "Spel2.exe"
        dest_exe = self.install_dir / "Spel2.exe"

        if is_patched(source_exe):
            logger.critical(
                "Source exe (%s) is somehow patched. You need to validate game files in steam and re-extract."
            )
            return

        logger.info("Copying exe from %s to %s", source_exe, dest_exe)
        shutil.copy2(source_exe, dest_exe)
        logger.info("Done")

    def pack(self):
        packs = [
            self.install_dir / "Mods" / exe.get()
            for exe in self.checkbox_vars
            if exe.get()
        ]
        thread = threading.Thread(target=self.repack_assets, args=(packs,))
        thread.start()

    def get_packs(self):
        pack_dirs = []
        overrides_dir = self.install_dir / "Mods" / "Overrides"
        if overrides_dir.exists():
            pack_dirs.append(overrides_dir.relative_to(self.install_dir / "Mods"))

        packs_dir = self.install_dir / "Mods" / "Packs"
        if packs_dir.exists():
            for dir_ in packs_dir.iterdir():
                if not dir_.is_dir():
                    continue
                pack_dirs.append(dir_.relative_to(self.install_dir / "Mods"))

        return pack_dirs

    def on_load(self):
        for child in self.frame.scrollable_frame.winfo_children():
            child.destroy()
        self.checkbox_vars.clear()

        for idx, exe in enumerate(self.get_packs()):
            str_var = tk.StringVar()
            item = tk.Checkbutton(
                self.frame.scrollable_frame,
                text=f" {exe}",
                image=self.default_icon,
                font=("Segoe UI", 12, "bold"),
                variable=str_var,
                onvalue=f"{exe}",
                offvalue="",
                compound="left",
            )
            self.checkbox_vars.append(str_var)
            item.grid(row=idx, column=0, pady=5, padx=5, sticky="w")

    @log_exception
    def repack_assets(self, packs):
        mods_dir = self.install_dir / MODS
        extract_dir = mods_dir / "Extracted"
        source_exe = extract_dir / "Spel2.exe"
        dest_exe = self.install_dir / "Spel2.exe"

        logger.info("Starting Pack of %s", source_exe)
        if is_patched(source_exe):
            logger.critical(
                "Source exe (%s) is somehow patched. You need to re-extract.", source_exe
            )
            return

        # If the destination isn't patched we want to check if it differs
        # from the source exe as new updates are a regular point of confusion
        # for users.
        if not is_patched(dest_exe):
            logger.info("Checking for new release...")
            logger.info("Hashing %s", source_exe)
            src_md5 = md5sum_path(source_exe)
            logger.info("Hashing %s", dest_exe)
            dest_md5 = md5sum_path(dest_exe)
            if src_md5 != dest_md5:
                logger.critical((
                    "%s appears to be a new version. You need to extract before packing again."
                ), dest_exe)
                return


        shutil.copy2(source_exe, dest_exe)

        with dest_exe.open("rb+") as dest_file:
            asset_store = AssetStore.load_from_file(dest_file)
            try:
                asset_store.repackage(
                    packs,
                    extract_dir,
                    mods_dir / ".compressed",
                )
            except MissingAsset as err:
                logger.error(
                    "Failed to find expected asset: %s. Unabled to proceed...", err
                )
                return

            patcher = Patcher(dest_file)
            patcher.patch_checksum()
            patcher.patch_release()
            logger.info("Repacking complete!")



import logging
import os
import queue
import random
import shutil
import subprocess
import threading
import tkinter as tk
import tkinter.messagebox as tkMessageBox
import webbrowser
from pathlib import Path
from tkinter import PhotoImage, ttk
from tkinter.scrolledtext import ScrolledText

import requests
from packaging import version
from PIL import Image, ImageTk

from modlunky2.assets.assets import AssetStore
from modlunky2.assets.constants import (
    EXTRACTED_DIR,
    FILEPATH_DIRS,
    OVERRIDES_DIR,
    PACKS_DIR,
)
from modlunky2.assets.exc import MissingAsset
from modlunky2.assets.patcher import Patcher
from modlunky2.constants import BASE_DIR, ROOT_DIR, IS_EXE
from modlunky2.ui.extract import ExtractTab
from modlunky2.ui.levels import LevelsTab
from modlunky2.ui.pack import PackTab
from modlunky2.ui.widgets import ConsoleWindow

logger = logging.getLogger("modlunky2")

cwd = os.getcwd()

MIN_WIDTH = 1024
MIN_HEIGHT = 800


def get_latest_version():
    try:
        return version.parse(
            requests.get(
                "https://api.github.com/repos/spelunky-fyi/modlunky2/releases/latest"
            ).json()["tag_name"]
        )
    except Exception:  # pylint: disable=broad-except
        return None


def get_current_version():
    with (ROOT_DIR / "VERSION").open() as version_file:
        return version.parse(version_file.read().strip())


class ModlunkyUI:
    def __init__(self, install_dir, beta=False):
        self.install_dir = install_dir
        self.beta = beta

        if IS_EXE:
            self.current_version = get_current_version()
            self.latest_version = get_latest_version()
            if self.latest_version is None or self.current_version is None:
                self.needs_update = False
            else:
                self.needs_update = self.current_version < self.latest_version
        else:
            self.current_version = None
            self.latest_version = None
            self.needs_update = False

        self._shutdown_handlers = []
        self._shutting_down = False

        self.root = tk.Tk(className="Modlunky2")  # Equilux Black
        self.root.title("Modlunky 2")
        self.root.geometry(f"{MIN_WIDTH}x{MIN_HEIGHT}")
        self.root.minsize(MIN_WIDTH, MIN_HEIGHT)
        self.root.grid_columnconfigure(0, weight=1)
        self.root.grid_rowconfigure(0, weight=0)
        self.root.grid_rowconfigure(1, weight=1)
        self.root.grid_rowconfigure(2, weight=0)
        # self.root.resizable(False, False)
        self.icon_png = PhotoImage(file=BASE_DIR / "static/images/icon.png")
        self.root.iconphoto(False, self.icon_png)

        if self.needs_update:
            update_button = ttk.Button(
                self.root, text="Update Modlunky2!", command=self.update
            )
            update_button.grid(column=0, row=0)

        # Handle shutting down cleanly
        self.root.protocol("WM_DELETE_WINDOW", self.quit)

        self.tabs = {}
        self.tab_control = ttk.Notebook(self.root)

        self.register_tab(
            "Pack Assets",
            PackTab(
                tab_control=self.tab_control,
                install_dir=install_dir,
            ),
        )
        self.register_tab(
            "Extract Assets",
            ExtractTab(
                tab_control=self.tab_control,
                install_dir=install_dir,
            ),
        )
        self.register_tab(
            "Levels",
            LevelsTab(
                tab_control=self.tab_control,
                install_dir=install_dir,
            ),
        )

        self.tab_control.bind("<<NotebookTabChanged>>", self.on_tab_change)
        self.tab_control.grid(column=0, row=1, sticky="nsew")

        self.console = ConsoleWindow(self.root)
        self.console.grid(column=0, row=2, sticky="ew")

    def update(self):
        webbrowser.open_new_tab(
            f"https://github.com/spelunky-fyi/modlunky2/releases/tag/{self.latest_version}"
        )

    def self_update(self):
        updater = Path(cwd + "/updater.exe")
        subprocess.call([updater])  # if file exists
        self.root.quit()
        self.root.destroy()

    def on_tab_change(self, event):
        tab = event.widget.tab("current")["text"]
        self.tabs[tab].on_load()

    def register_tab(self, name, obj):
        self.tabs[name] = obj
        self.tab_control.add(obj, text=name)

    def quit(self):
        if self._shutting_down:
            return

        self._shutting_down = True
        logger.info("Shutting Down.")
        for handler in self._shutdown_handlers:
            handler()

        self.root.quit()
        self.root.destroy()

    def register_shutdown_handler(self, func):
        self._shutdown_handlers.append(func)

    def mainloop(self):
        try:
            self.root.mainloop()
        except KeyboardInterrupt:
            self.quit()



import logging
import shutil
import threading
import tkinter as tk
from pathlib import Path
from tkinter import ttk

from modlunky2.assets.assets import AssetStore
from modlunky2.assets.constants import (
    EXTRACTED_DIR,
    FILEPATH_DIRS,
    OVERRIDES_DIR,
    PACKS_DIR,
)
from modlunky2.ui.utils import is_patched, log_exception
from modlunky2.ui.widgets import Tab, ToolTip

logger = logging.getLogger("modlunky2")


MODS = Path("Mods")

TOP_LEVEL_DIRS = [EXTRACTED_DIR, PACKS_DIR, OVERRIDES_DIR]


class ExtractTab(Tab):
    def __init__(self, tab_control, install_dir, *args, **kwargs):
        super().__init__(tab_control, *args, **kwargs)
        self.tab_control = tab_control
        self.install_dir = install_dir

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, minsize=60)

        self.top_frame = ttk.Frame(self)
        self.top_frame.rowconfigure(0, weight=1)
        self.top_frame.columnconfigure(0, weight=1)
        self.top_frame.columnconfigure(1, minsize=250)
        self.top_frame.grid(row=0, column=0, sticky="nswe")

        self.exe_frame = ttk.LabelFrame(self.top_frame, text="Select exe to Extract")
        self.exe_frame.grid(row=0, column=0, pady=5, padx=5, sticky="nswe")
        self.exe_frame.rowconfigure(0, weight=1)
        self.exe_frame.columnconfigure(0, weight=1)
        self.exe_frame.columnconfigure(1)

        self.list_box = tk.Listbox(self.exe_frame)
        self.list_box.grid(row=0, column=0, sticky="nswe")
        self.scrollbar = ttk.Scrollbar(self.exe_frame)
        self.scrollbar.grid(row=0, column=1, sticky="nes")

        self.config_frame = ttk.LabelFrame(self.top_frame, text="Config")
        self.config_frame.grid(row=0, column=1, pady=5, padx=5, sticky="nswe")

        self.recompress = tk.BooleanVar()
        self.recompress.set(True)
        self.checkbox_recompress = tk.Checkbutton(
            self.config_frame,
            text='Recompress', variable=self.recompress, onvalue=True, offvalue=False,
        )
        self.checkbox_recompress.grid(row=0, sticky="nw")
        ToolTip(self.checkbox_recompress, (
            "Recompress assets to speed up futuring packing.\n"
            "Not necessary if you just want the extracted assets."
        ))

        self.create_entity = tk.BooleanVar()
        self.create_entity.set(True)
        self.checkbox_create_entity = tk.Checkbutton(
            self.config_frame,
            text='Create Entity Sprites', variable=self.create_entity, onvalue=True, offvalue=False,
        )
        self.checkbox_create_entity.grid(row=1, sticky="nw")
        ToolTip(self.checkbox_create_entity, (
            "Create merged entity spritesheets. These provide a simpler\n"
            "interface to some entity mods."
        ))

        self.list_box.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.list_box.yview)

        self.button_extract = ttk.Button(self, text="Extract", command=self.extract)
        self.button_extract.grid(row=1, column=0, pady=5, padx=5, sticky="nswe")
        ToolTip(self.button_extract, (
            "Extract assets from EXE."
        ))

    def extract(self):
        idx = self.list_box.curselection()
        if not idx:
            return

        selected_exe = self.list_box.get(idx)
        thread = threading.Thread(
            target=self.extract_assets,
            args=(selected_exe, self.recompress.get(), self.create_entity.get()))
        thread.start()

    def get_exes(self):
        exes = []
        # Don't recurse forever. 3 levels should be enough
        exes.extend(self.install_dir.glob("*.exe"))
        exes.extend(self.install_dir.glob("*/*.exe"))
        exes.extend(self.install_dir.glob("*/*/*.exe"))
        return [
            exe.relative_to(self.install_dir)
            for exe in exes
            # Exclude modlunky2 which is likely in the install directory
            if exe.name not in ["modlunky2.exe"]
        ]

    def on_load(self):
        self.list_box.delete(0, tk.END)
        for exe in self.get_exes():
            self.list_box.insert(tk.END, str(exe))

    @log_exception
    def extract_assets(self, target, recompress, create_entity_sheets):

        exe_filename = self.install_dir / target

        if is_patched(exe_filename):
            logger.critical((
                "%s is a patched exe. Can't extract. You should Restore Exe"
                " or validate game files to get a clean exe before Extracting."
            ), exe_filename)
            return

        mods_dir = self.install_dir / MODS

        for dir_ in TOP_LEVEL_DIRS:
            (mods_dir / dir_).mkdir(parents=True, exist_ok=True)

        for dir_ in FILEPATH_DIRS:
            (mods_dir / EXTRACTED_DIR / dir_).mkdir(parents=True, exist_ok=True)
            (mods_dir / ".compressed" / EXTRACTED_DIR / dir_).mkdir(
                parents=True, exist_ok=True
            )

        with exe_filename.open("rb") as exe:
            asset_store = AssetStore.load_from_file(exe)
            unextracted = asset_store.extract(
                mods_dir / EXTRACTED_DIR,
                mods_dir / ".compressed" / EXTRACTED_DIR,
                recompress=recompress,
                create_entity_sheets=create_entity_sheets,
            )

        for asset in unextracted:
            logger.warning("Un-extracted Asset %s", asset.asset_block)

        dest = mods_dir / EXTRACTED_DIR / "Spel2.exe"
        if exe_filename != dest:
            logger.info("Backing up exe to %s", dest)
            shutil.copy2(exe_filename, dest)

        logger.info("Extraction complete!")



from .base_classes import AbstractFloorSheet

__all__ = [
    "CaveFloorSheet",
    "VolcanaFloorSheet",
    "JungleFloorSheet",
    "TidePoolFloorSheet",
    "TempleFloorSheet",
    "IceCavesFloorSheet",
    "SurfaceFloorSheet",
    "SunkenCityFloorSheet",
    "BabylonFloorSheet",
    "EggplantFloorSheet",
]


class CaveFloorSheet(AbstractFloorSheet):
    biome_name = "cave"


class VolcanaFloorSheet(AbstractFloorSheet):
    biome_name = "volcano"


class JungleFloorSheet(AbstractFloorSheet):
    biome_name = "jungle"


class TidePoolFloorSheet(AbstractFloorSheet):
    biome_name = "tidepool"


class TempleFloorSheet(AbstractFloorSheet):
    biome_name = "temple"


class IceCavesFloorSheet(AbstractFloorSheet):
    biome_name = "ice"


class BabylonFloorSheet(AbstractFloorSheet):
    biome_name = "babylon"


class SunkenCityFloorSheet(AbstractFloorSheet):
    biome_name = "sunken"
    _additional_chunks = {"bigspear_trap": (8, 9, 10, 10)}


class EggplantFloorSheet(AbstractFloorSheet):
    biome_name = "eggplant"


class SurfaceFloorSheet(AbstractFloorSheet):
    biome_name = "surface"



from pathlib import Path

from .base_classes import BaseSpriteLoader


class JournalPlaceSheet(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/journal_entry_place.png")
    _chunk_size = 320
    _chunk_map = {
        "journal_dwelling": (0, 0, 2, 1),
        "journal_jungle": (2, 0, 4, 1),
        "journal_volcana": (4, 0, 6, 1),
        "journal_olmecs_lair": (6, 0, 8, 1),
        "journal_tide_pool": (0, 1, 2, 2),
        "journal_temple_of_anubis": (2, 1, 4, 2),
        "journal_ice_caves": (4, 1, 6, 2),
        "journal_neo_babylon": (6, 1, 8, 2),
        "journal_sunken_city": (0, 2, 2, 3),
        "journal_cosmic_ocean": (2, 2, 4, 3),
        "journal_city_of_gold": (4, 2, 6, 3),
        "journal_duat": (6, 2, 8, 3),
        "journal_abzu": (0, 3, 2, 4),
        "journal_tiamats_throne": (2, 3, 4, 4),
        "journal_eggplant_world": (4, 3, 6, 4),
        "journal_hunduns_hideaway": (6, 3, 8, 4),
    }



from pathlib import Path

from ..base_classes.base_sprite_merger import BaseSpriteMerger
from ..monsters.pets import Pets
from ..journal_mons import JournalMonsterSheet
from ..util import chunks_from_animation


class MontySpriteMerger(BaseSpriteMerger):
    _target_sprite_sheet_path = Path("Data/Textures/Entities/monty_full.png")
    _grid_hint_size = 8
    _origin_map = {
        Pets: {
            **chunks_from_animation("monty_row_0_", (0, 0, 1, 1), 12),
            **chunks_from_animation("monty_row_1_", (0, 1, 1, 2), 10),
            **chunks_from_animation("monty_row_2_", (0, 2, 1, 3), 12),
            **chunks_from_animation("monty_row_3_", (0, 3, 1, 4), 12),
        },
        JournalMonsterSheet: {"journal_monty": (0, 0, 1, 1)},
    }


class PercySpriteMerger(BaseSpriteMerger):
    _target_sprite_sheet_path = Path("Data/Textures/Entities/percy_full.png")
    _grid_hint_size = 8
    _origin_map = {
        Pets: {
            **chunks_from_animation("percy_row_0_", (0, 0, 1, 1), 12),
            **chunks_from_animation("percy_row_1_", (0, 1, 1, 2), 10),
            **chunks_from_animation("percy_row_2_", (0, 2, 1, 3), 12),
            **chunks_from_animation("percy_row_3_", (0, 3, 1, 4), 12),
        },
        JournalMonsterSheet: {"journal_percy": (0, 0, 1, 1)},
    }


class PoochiSpriteMerger(BaseSpriteMerger):
    _target_sprite_sheet_path = Path("Data/Textures/Entities/poochi_full.png")
    _grid_hint_size = 8
    _origin_map = {
        Pets: {
            **chunks_from_animation("poochi_row_0_", (0, 0, 1, 1), 12),
            **chunks_from_animation("poochi_row_1_", (0, 1, 1, 2), 10),
            **chunks_from_animation("poochi_row_2_", (0, 2, 1, 3), 12),
            **chunks_from_animation("poochi_row_3_", (0, 3, 1, 4), 12),
        },
        JournalMonsterSheet: {"journal_poochi": (0, 0, 1, 1)},
    }



from .level_file import LevelFile
from .level_chances import LevelChance
from .level_templates import LevelTemplate
from .monster_chances import MonsterChance
from .level_settings import LevelSetting
from .tile_codes import TileCode
